import pytest

from dlms_cosem import cosem, enumerations
from dlms_cosem.protocol import xdlms


class TestSetRequestNormal:
    def test_transform_bytes(self):
        data = b"\xc1\x01\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        request = xdlms.SetRequestNormal(
            cosem_attribute=cosem.CosemAttribute(
                interface=enumerations.CosemInterface.CLOCK,
                instance=cosem.Obis(a=0, b=0, c=1, d=0, e=0, f=255),
                attribute=2,
            ),
            data=b"\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00",
            access_selection=None,
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=1, confirmed=True, high_priority=True
            ),
        )
        assert data == request.to_bytes()
        assert request == xdlms.SetRequestNormal.from_bytes(data)

    def test_wrong_tag_raises_value_error(self):
        data = b"\xc2\x01\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        with pytest.raises(ValueError):
            xdlms.SetRequestNormal.from_bytes(data)

    def test_wrong_type_raises_value_error(self):

        data = b"\xc1\x02\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        with pytest.raises(ValueError):
            xdlms.SetRequestNormal.from_bytes(data)


class TestSetRequestFactory:
    def test_set_request_normal(self):
        data = b"\xc1\x01\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        request = xdlms.SetRequestFactory.from_bytes(data)
        assert isinstance(request, xdlms.SetRequestNormal)

    def test_wrong_tag_raises_value_error(self):
        data = b"\xc2\x01\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        with pytest.raises(ValueError):
            xdlms.SetRequestFactory.from_bytes(data)

    def test_request_with_first_block_raises_not_implemented_error(self):
        data = b"\xc1\x02\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetRequestFactory.from_bytes(data)

    def test_set_request_with_block_raises_not_implemented_error(self):
        data = b"\xc1\x03\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetRequestFactory.from_bytes(data)

    def test_set_with_list_raises_not_implemented_error(self):
        data = b"\xc1\x04\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetRequestFactory.from_bytes(data)

    def test_set_request_with_list_first_block_raises_not_implemented_block(self):
        data = b"\xc1\x05\xc1\x00\x08\x00\x00\x01\x00\x00\xff\x02\x00\t\x0c\x07\xe5\x01\x18\xff\x0e09P\xff\xc4\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetRequestFactory.from_bytes(data)


class TestSetResponseNormal:
    def test_transform_bytes(self):
        data = b"\xc5\x01\xc1\x00"
        response = xdlms.SetResponseNormal(
            result=enumerations.DataAccessResult.SUCCESS,
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=1, confirmed=True, high_priority=True
            ),
        )
        assert data == response.to_bytes()
        assert response == xdlms.SetResponseNormal.from_bytes(data)

    def test_wrong_tag_raises_value_error(self):
        data = b"\xc6\x01\xc1\x00"
        with pytest.raises(ValueError):
            xdlms.SetRequestNormal.from_bytes(data)

    def test_wrong_type_raises_value_error(self):
        data = b"\xc5\x02\xc1\x00"
        with pytest.raises(ValueError):
            xdlms.SetRequestNormal.from_bytes(data)


class TestSetResponseFactory:
    def test_set_response_normal(self):
        data = b"\xc5\x01\xc1\x00"
        request = xdlms.SetResponseFactory.from_bytes(data)
        assert isinstance(request, xdlms.SetResponseNormal)

    def test_wrong_tag_raises_value_error(self):
        data = b"\xc6\x01\xc1\x00"
        with pytest.raises(ValueError):
            xdlms.SetResponseFactory.from_bytes(data)

    def test_set_response_with_block_raises_not_implemented_error(self):
        data = b"\xc5\x02\xc1\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetResponseFactory.from_bytes(data)

    def test_set_response_last_block_raises_not_implemented_error(self):
        data = b"\xc5\x03\xc1\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetResponseFactory.from_bytes(data)

    def test_set_response_last_block_with_list_raises_not_implemented_error(self):
        data = b"\xc5\x04\xc1\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetResponseFactory.from_bytes(data)

    def test_set_response_with_list_raises_not_implemented_error(self):
        data = b"\xc5\x05\xc1\x00"
        with pytest.raises(NotImplementedError):
            xdlms.SetResponseFactory.from_bytes(data)



import os
from typing import *

import attr
import attrs
import structlog

from dlms_cosem import enumerations as enums
from dlms_cosem import exceptions, security
from dlms_cosem import state as dlms_state
from dlms_cosem import utils
from dlms_cosem.security import AuthenticationMethodManager, NoSecurityAuthentication
from dlms_cosem.protocol import acse, xdlms
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.conformance import Conformance

LOG = structlog.get_logger()


def default_system_title() -> bytes:
    """A non FLAG registed id + 5 random bytes"""
    return b"uti" + os.urandom(5)


class XDlmsApduFactory:
    """
    A factory to return the correct APDU depending from the tag.
    """

    APDU_MAP = {
        1: xdlms.InitiateRequest,
        8: xdlms.InitiateResponse,
        14: xdlms.ConfirmedServiceError,
        15: xdlms.DataNotification,
        33: xdlms.GlobalCipherInitiateRequest,
        40: xdlms.GlobalCipherInitiateResponse,
        216: xdlms.ExceptionResponse,
        219: xdlms.GeneralGlobalCipher,
        # ACSE APDUs:
        96: acse.ApplicationAssociationRequest,
        97: acse.ApplicationAssociationResponse,
        98: acse.ReleaseRequest,
        99: acse.ReleaseResponse,
        192: xdlms.GetRequestFactory,
        193: xdlms.SetRequestFactory,
        195: xdlms.ActionRequestFactory,
        196: xdlms.GetResponseFactory,
        197: xdlms.SetResponseFactory,
        199: xdlms.ActionResponseFactory,
    }

    @classmethod
    def apdu_from_bytes(cls, apdu_bytes):
        tag = apdu_bytes[0]

        try:
            apdu_class = cls.APDU_MAP[tag]
        except KeyError as e:
            raise KeyError(f"Tag {tag!r} is not available in DLMS APDU Factory") from e
        return apdu_class.from_bytes(apdu_bytes)


def make_client_to_server_challenge(length: int = 8) -> bytes:
    """
    Return a valid challenge depending on the authentocation method.
    """
    if 8 <= length <= 64:
        return os.urandom(length)
    else:
        raise ValueError(
            f"Client to server challenge must be between 8 and 64 bytes. Got {length}"
        )


def make_conformance(encryption_key: Optional[bytes], use_block_transfer: bool):
    """
    Return a default conformance with general_protection set if a
    encryption key is passed.
    """
    return Conformance(
        general_protection=bool(encryption_key),
        general_block_transfer=True,
        delta_value_encoding=False,
        attribute_0_supported_with_set=False,
        priority_management_supported=True,
        attribute_0_supported_with_get=False,
        block_transfer_with_get_or_read=True,
        block_transfer_with_set_or_write=False,
        block_transfer_with_action=False,
        multiple_references=True,
        data_notification=False,
        access=True,
        get=True,
        set=True,
        selective_access=True,
        event_notification=True,
        action=True,
    )


class ProtectionError(Exception):
    """Unable to perform cryptographic function"""


@attr.s(auto_attribs=True)
class DlmsConnectionSettings:
    """
    Class to hold values and settings to handle different quirks of different dlms
    server implementations and manufacturers specific irregularity.
    """

    # If false there should be no RLRQ and RLRE to release an association. It is fine to disconnect the lower layer
    # directly
    use_rlrq_rlre: bool = attr.ib(default=True)
    # In Pietro Fiorentini local communication over HDLC the system title in GeneralGlobalCiphering is omitted.
    empty_system_title_in_general_glo_ciphering: bool = attr.ib(default=False)


@attr.s(auto_attribs=True)
class DlmsConnection:
    """
    A DLMS connection.
    """

    authentication: AuthenticationMethodManager

    # Client system title can be any combination of 8 bytes.
    # But is should not be the same as the metering the connection is set up too.
    client_system_title: bytes = attr.ib(
        converter=attr.converters.default_if_none(
            factory=default_system_title
        )  # type: ignore
    )

    global_encryption_key: Optional[bytes] = attr.ib(default=None)
    global_authentication_key: Optional[bytes] = attr.ib(default=None)

    # The security suite just indicates what cryptographical functions are available to
    # the meter
    security_suite: int = attr.ib(default=0)

    # Meter system title is usually unknown at the start of the connection. And it is
    # only needed if the connection is ciphered. In the AARE the meter will respond with
    # its system title
    meter_system_title: Optional[bytes] = attr.ib(default=None)

    # # Meter authentication method.
    # # TODO: this should not be set to None. Should not be optional.
    # authentication_method: Optional[enums.AuthenticationMechanism] = attr.ib(
    #     default=None
    # )
    # # Low Level Security (LLS) password
    # password: Optional[bytes] = attr.ib(default=None)
    #
    # # HLS challenge length.
    # challenge_length: int = attr.ib(default=32)
    #
    # # client_to_meter_challenge is generated automatically with a random seed
    # # depending on the HLS setup.
    # client_to_meter_challenge: bytes = attr.ib(
    #     init=False,
    #     default=attr.Factory(
    #         lambda self: make_client_to_server_challenge(self.challenge_length),
    #         takes_self=True,
    #     ),
    # )
    meter_to_client_challenge: Optional[bytes] = attr.ib(default=None, init=False)

    # To keep track of invocation counters used by the meter. If a request is received
    # with an invocation counter smaller than the current registered the message should
    # be rejected. If the meter receives a message with a with the same or lower
    # invocation counter as the last received message it will reject the message. So
    # after every use of the apdu protection the client_invocation_counter needs to be
    # increased
    client_invocation_counter: int = attr.ib(default=0)
    meter_invocation_counter: int = attr.ib(default=0)

    # Using dedicated ciphering increase security since we only use the global key for
    # association setup. This also saves a bit on the global invocation counter as once
    # it hits the maximum value the global key needs to be exchanged.
    use_dedicated_ciphering: bool = attr.ib(default=False)
    # Dedicated key will be generated on each association
    global_dedicated_key: Optional[bytes] = attr.ib(default=None)
    # the dedicated invocation_counter will be reset on each new dedicated_key.
    dedicated_invocation_counter: int = attr.ib(init=False, default=0)

    # not supported yet
    use_block_transfer: bool = attr.ib(default=False)

    # the max pdu size controls when we need to use block transfer. If the message is
    # larger than max_pdu_size we automatically use the general block service.
    # Unless it is not suppoeted in conformance. Then raise error.
    max_pdu_size: int = attr.ib(default=65535)

    # When a connection is preestablished we wont allow any ACSE adpus.
    is_pre_established: bool = attr.ib(default=False)

    buffer: bytearray = attr.ib(init=False, factory=bytearray)
    state: dlms_state.DlmsConnectionState = attr.ib(
        factory=dlms_state.DlmsConnectionState
    )

    conformance: Conformance = attr.ib(
        default=attr.Factory(
            lambda self: make_conformance(
                self.global_encryption_key, self.use_block_transfer
            ),
            takes_self=True,
        )
    )

    settings: DlmsConnectionSettings = attr.ib(
        default=DlmsConnectionSettings(),
        converter=attr.converters.default_if_none(factory=DlmsConnectionSettings),
    )

    @classmethod
    def with_pre_established_association(
        cls,
        conformance: Conformance,
        max_pdu_size: int = 65535,
        global_encryption_key: Optional[bytes] = None,
        global_authentication_key: Optional[bytes] = None,
        client_invocation_counter: Optional[int] = None,
        meter_invocation_counter: Optional[int] = None,
        client_system_title: Optional[bytes] = None,
        meter_system_title: Optional[bytes] = None,
    ):
        """
        A pre-established association does not need the ACSE APDUs. It is
        predetermined what access the client have.
        """
        return cls(
            client_system_title=client_system_title,
            meter_system_title=meter_system_title,
            global_encryption_key=global_encryption_key,
            global_authentication_key=global_authentication_key,
            meter_invocation_counter=meter_invocation_counter,
            client_invocation_counter=client_invocation_counter,
            # Moves the state into ready.
            state=dlms_state.DlmsConnectionState(current_state=dlms_state.READY),
            is_pre_established=True,
            conformance=conformance,
            max_pdu_size=max_pdu_size,
            authentication=NoSecurityAuthentication(),
        )

    @property
    def security_control(self) -> security.SecurityControlField:
        """
        The security control field is used in encryption/decryption of data. It also
        follows the protected apdus to indicate what kind of protections they have.
        """
        return security.SecurityControlField(
            self.security_suite,
            encrypted=bool(self.global_encryption_key),
            authenticated=bool(self.global_authentication_key),
            broadcast_key=False,
        )

    def send(self, event) -> bytes:
        """
        Returns the bytes to be sent over the connection and changes the state
        depending on event sent.
        """

        if self.is_pre_established:
            # When we are in a pre established association state starts as READY.
            # Only invalid state change is to send the ReleaseRequestApdu. But it is not
            # possible to close a pre-established association.
            if isinstance(
                event, (acse.ReleaseRequest, acse.ApplicationAssociationRequest)
            ):
                raise exceptions.PreEstablishedAssociationError(
                    f"It is not allowed to send a {type(event)} when the association is"
                    f"pre-established "
                )

        if not self.settings.use_rlrq_rlre and isinstance(event, acse.ReleaseRequest):
            # Client has issued a release request but the connection is not using them
            # Connection states goes to NO_ASSOCIATION directly

            LOG.info(
                "Stopped ReleaseRequest as settings.use_rlrq_rlre is False. Ending association",
                settings=self.settings,
                rlrq=event,
            )
            self.state.process_event(dlms_state.EndAssociation())
            raise exceptions.NoRlrqRlreError(
                "Connection settings does not allow ReleaseRequest and ReleaseResponse"
            )

        self.state.process_event(event)
        LOG.debug(f"Preparing to send DLMS Request", request=event)

        if self.use_protection:
            event = self.protect(event)

        # if self.use_blocks:
        #    blocks = self.make_blocks(event)
        #    # TODO: How to handle the subcase of sending blocks?

        LOG.info(f"Sending DLMS Request", request=event)

        out = event.to_bytes()

        if len(out) > self.max_pdu_size:
            raise exceptions.LocalDlmsProtocolError(
                f"PDU size too big. Max PDU size for association is {self.max_pdu_size} "
                f"bytes. PDU to be sent is {len(out)}"
            )
        return out

    def receive_data(self, data: bytes):
        """
        Add data into the receive buffer.
        After this you could call next_event
        """
        if data:
            LOG.debug(f"Adding data to buffer", data=data)
            self.buffer += data

    def next_event(self):
        """
        Will parse the buffer into an APDU. In lower levels we need the case to get more
        data. But this is not needed in the DLMS connections as it is the lower layers
        responsibility to make sure the data is complete before handing the control
        back to the dlms-layer. In the HDLC case the data is complete when we get the
        last unsegmented information frame. And in the IP case the length is known from
        the IP wrapper element so it is possible to can keep on trying until all data
        is received.
        """

        apdu = XDlmsApduFactory.apdu_from_bytes(self.buffer)

        LOG.info("Received DLMS Response", response=apdu)

        if isinstance(apdu, acse.ApplicationAssociationResponse):
            # To be able to run the decryption we need to know some things about the
            # meter and that has to be extracted first
            self.update_meter_info(apdu)

        if self.use_protection:
            apdu = self.unprotect(apdu)
            LOG.info("Deciphered DLMS Response", response=apdu)

        if self.is_pre_established:
            if isinstance(
                apdu,
                (acse.ApplicationAssociationResponse, acse.ReleaseResponse),
            ):
                raise exceptions.PreEstablishedAssociationError(
                    f"Received a {apdu.__class__.__name__}. In a pre-established "
                    f"association it is not possible to handle ACSE services."
                )

        self.state.process_event(apdu)
        self.clear_buffer()

        if isinstance(apdu, acse.ApplicationAssociationResponse):
            self.update_negotiated_parameters(apdu)

            if apdu.result in [
                enums.AssociationResult.REJECTED_PERMANENT,
                enums.AssociationResult.REJECTED_TRANSIENT,
            ]:
                # reset the association on a reject
                self.state.process_event(dlms_state.RejectAssociation())
                return apdu

            # we need to start the HLS auth.
            if apdu.authentication:
                if apdu.authentication >= enums.AuthenticationMechanism.HLS:
                    self.state.process_event(dlms_state.HlsStart())

        # Handle HLS verification
        if self.state.current_state == dlms_state.HLS_DONE:
            if isinstance(apdu, xdlms.ActionResponseNormalWithData):
                if apdu.status != enums.ActionResultStatus.SUCCESS:
                    self.state.process_event(dlms_state.HlsFailed())
                if self.authentication.hls_meter_data_is_valid(
                    utils.parse_as_dlms_data(apdu.data), self
                ):
                    self.state.process_event(dlms_state.HlsSuccess())
                else:
                    self.state.process_event(dlms_state.HlsFailed())
            elif isinstance(
                apdu, (xdlms.ActionResponseNormalWithError, xdlms.ActionResponseNormal)
            ):
                self.state.process_event(dlms_state.HlsFailed())

            else:
                raise exceptions.LocalDlmsProtocolError(
                    "Received a non Action response when in HLS DONE"
                )
        return apdu

    def clear_buffer(self):
        self.buffer = bytearray()

    @property
    def use_protection(self) -> bool:
        """
        If the Association is such that APDUs should be protected.
        :return:
        """
        if (
            self.global_encryption_key is not None
            or self.global_authentication_key is not None
        ):
            return True
        else:
            return False

    def protect(self, event) -> Any:
        """
        Will apply the correct protection to apdus depending on the security context
        """

        LOG.info(f"Ciphering DLMS Request", apdu=event)

        # ASCE have different rules about protection
        if isinstance(event, (acse.ApplicationAssociationRequest, acse.ReleaseRequest)):
            if event.user_information:
                ciphered_text, ic = self.encrypt(
                    event.user_information.content.to_bytes()
                )

                event.user_information = acse.UserInformation(
                    content=xdlms.GlobalCipherInitiateRequest(
                        security_control=self.security_control,
                        invocation_counter=ic,
                        ciphered_text=ciphered_text,
                    )
                )

        # XDLMS apdus should be protected with general-glo-ciphering
        elif isinstance(event, AbstractXDlmsApdu):
            ciphered_text, ic = self.encrypt(event.to_bytes())

            if self.settings.empty_system_title_in_general_glo_ciphering:
                system_title = None
            else:
                system_title = self.client_system_title

            event = xdlms.GeneralGlobalCipher(
                system_title=system_title,
                security_control=self.security_control,
                invocation_counter=ic,
                ciphered_text=ciphered_text,
            )
        else:
            raise RuntimeError(f"Unable to handle ecryption/protection of {event}")

        return event

    def encrypt(self, plain_text: bytes) -> Tuple[bytes, int]:
        """
        Encrypts plain bytes according to the current association and connection.
        Returns the ciphered text and the invocation counter used with the ciphered text.
        It also increases the internal client invocation counter to make sure a new
        invocation counter is used at every encryption call.
        """
        if not self.global_encryption_key:
            raise ProtectionError(
                "Unable to encrypt plain text. Missing global_encryption_key"
            )
        if not self.global_authentication_key:
            raise ProtectionError(
                "Unable to encrypt plain text. Missing global_authentication_key"
            )

        invocation_counter = self.client_invocation_counter

        ciphered_text = security.encrypt(
            self.security_control,
            system_title=self.client_system_title,
            invocation_counter=invocation_counter,
            key=self.global_encryption_key,
            auth_key=self.global_authentication_key,
            plain_text=plain_text,
        )

        # updated the client_invocation_counter
        self.client_invocation_counter += 1

        return ciphered_text, invocation_counter

    def decrypt(
        self,
        ciphered_text: bytes,
    ):
        """
        Encrypts ciphered bytes according to the current association and connection.
        In the case of AARE we have not had the opportunity to
        """

        if not self.global_encryption_key:
            raise ProtectionError(
                "Unable to decrypt ciphered text. Missing global_encryption_key"
            )
        if not self.global_authentication_key:
            raise ProtectionError(
                "Unable to decrypt ciphered text. Missing global_authentication_key"
            )
        if not self.meter_system_title:
            raise ProtectionError(
                "Unable to decrypt ciphered text. Have not received the meters system title."
            )

        return security.decrypt(
            self.security_control,
            system_title=self.meter_system_title,
            invocation_counter=self.meter_invocation_counter,
            key=self.global_encryption_key,
            auth_key=self.global_authentication_key,
            cipher_text=ciphered_text,
        )

    def unprotect(self, event):
        """
        Removes protection from APDUs and return a new the unprotected version.
        It is necessary to remove update the connections meter invocation counter
        before encryption so that the connection uses the correct invocation counter
        when trying to decrypt.
        """

        if isinstance(
            event, (acse.ApplicationAssociationResponse, acse.ReleaseResponse)
        ):
            if event.user_information:
                if isinstance(
                    event.user_information.content,
                    xdlms.GlobalCipherInitiateResponse,
                ):
                    self.update_meter_invocation_counter(
                        event.user_information.content.invocation_counter
                    )
                    plain_text = self.decrypt(
                        event.user_information.content.ciphered_text
                    )
                    # Replace the ciphered InitiateResponse with the decrypted.
                    event.user_information.content = xdlms.InitiateResponse.from_bytes(
                        plain_text
                    )

        elif isinstance(event, xdlms.GeneralGlobalCipher):
            self.update_meter_invocation_counter(event.invocation_counter)
            plain_text = self.decrypt(event.ciphered_text)
            return XDlmsApduFactory.apdu_from_bytes(plain_text)

        return event

    @property
    def use_blocks(self) -> bool:
        """
        If the event should be sent via GlobalBlockTransfer
        """
        return False

    def make_blocks(self, event) -> List[Any]:
        """
        Will split an APDU in blocks
        """
        return event

    def get_aarq(self) -> acse.ApplicationAssociationRequest:
        """
        Returns an AARQ with the appropriate information for setting up the
        connection.
        """
        if self.global_encryption_key:
            ciphered_apdus = True
        else:
            ciphered_apdus = False

        initiate_request = xdlms.InitiateRequest(
            proposed_conformance=self.conformance,
            client_max_receive_pdu_size=self.max_pdu_size,
        )

        return acse.ApplicationAssociationRequest(
            ciphered=ciphered_apdus,
            system_title=self.client_system_title,
            authentication=self.authentication.authentication_method,
            authentication_value=self.authentication.get_calling_authentication_value(),
            user_information=acse.UserInformation(content=initiate_request),
        )

    def get_rlrq(self) -> acse.ReleaseRequest:
        """
        Returns a ReleaseRequestApdu to release the current association if one should be used.
        """

        initiate_request = xdlms.InitiateRequest(
            proposed_conformance=self.conformance,
            client_max_receive_pdu_size=self.max_pdu_size,
        )

        return acse.ReleaseRequest(
            reason=enums.ReleaseRequestReason.NORMAL,
            user_information=acse.UserInformation(content=initiate_request),
        )

    def update_negotiated_parameters(
        self, aare: acse.ApplicationAssociationResponse
    ) -> None:
        """
        When an AARE is received we need to update the connection to the negotiated
        parameters from the server (meter)
        """

        if aare.user_information:
            if isinstance(aare.user_information.content, xdlms.InitiateResponse):
                self.conformance = aare.user_information.content.negotiated_conformance
                self.max_pdu_size = (
                    aare.user_information.content.server_max_receive_pdu_size
                )

    def update_meter_invocation_counter(self, received_invocation_counter: int) -> None:
        """
        The received invocation counter must be larger than the last one we registered.
        """
        if received_invocation_counter <= self.meter_invocation_counter:
            raise exceptions.LocalDlmsProtocolError(
                "Received invocation counter is not larger than the previous "
                "received one. "
            )
        self.meter_invocation_counter = received_invocation_counter

    def update_meter_info(self, aare: acse.ApplicationAssociationResponse) -> None:
        self.meter_system_title = aare.system_title
        # if aare.authentication != self.authentication.authentication_method:
        #     raise RuntimeError(
        #         f"Meter adverticed a different authentication method than client. "
        #         f"Client authentication method: "
        #         f"{self.authentication.authentication_method}, "
        #         f"Meter authenticstion method: {aare.authentication}"
        #     )
        self.authentication_method = aare.authentication
        self.meter_to_client_challenge = aare.authentication_value



import pytest

from dlms_cosem.utils import parse_as_dlms_data
from dlms_cosem import dlms_data

def test_parse_data_from_kamstrup_han_port():
    data = b"\x02\x19\n\x0eKamstrup_V0001\t\x06\x01\x01\x00\x00\x05\xff\n\x105706567196382485\t\x06\x01\x01`\x01\x01\xff\n\x126841131BN143101090\t\x06\x01\x01\x01\x07\x00\xff\x06\x00\x00\t&\t\x06\x01\x01\x02\x07\x00\xff\x06\x00\x00\x00\x00\t\x06\x01\x01\x03\x07\x00\xff\x06\x00\x00\x00\x00\t\x06\x01\x01\x04\x07\x00\xff\x06\x00\x00\x00\xdf\t\x06\x01\x01\x1f\x07\x00\xff\x06\x00\x00\x00\\\t\x06\x01\x013\x07\x00\xff\x06\x00\x00\x00\x8d\t\x06\x01\x01G\x07\x00\xff\x06\x00\x00\x03r\t\x06\x01\x01 \x07\x00\xff\x12\x00\xe6\t\x06\x01\x014\x07\x00\xff\x12\x00\xe6\t\x06\x01\x01H\x07\x00\xff\x12\x00\xe4"

    parsed = parse_as_dlms_data(data)

    assert len(parsed) == 25


class TestVisibleString:

    parameter_data = [
        (b"\n\x0eKamstrup_V0001", "Kamstrup_V0001"),
        (b"\n\x105706567196382485", "5706567196382485"),
        (b"\n\x126841131BN143101090", "6841131BN143101090"),
    ]

    @pytest.mark.parametrize("encoded,decoded", parameter_data)
    def test_parse_data(self, encoded, decoded):
        parsed = parse_as_dlms_data(encoded)
        assert parsed == decoded

    @pytest.mark.parametrize("encoded,decoded", parameter_data)
    def test_encode_data(self, encoded, decoded):
        obj = dlms_data.VisibleStringData(value=decoded)

        assert obj.to_bytes() == encoded



import pytest

from dlms_cosem.protocol.xdlms import Conformance

# Example code found encoding found in DLMS Green Book v10 page:437


@pytest.mark.parametrize(
    "conformance,encoded",
    [
        (
            Conformance(
                priority_management_supported=True,
                attribute_0_supported_with_get=True,
                block_transfer_with_action=True,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=True,
                multiple_references=True,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            b"\x00\x00\x7e\x1f",
        ),
        (
            Conformance(
                priority_management_supported=True,
                block_transfer_with_get_or_read=True,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            b"\x00\x00\x50\x1f",
        ),
    ],
)
def test_conformance(conformance: Conformance, encoded: bytes):
    assert conformance.to_bytes() == encoded
    assert Conformance.from_bytes(encoded) == conformance



from functools import partial
from typing import *

import attr

from dlms_cosem import enumerations
from dlms_cosem.ber import BER
from dlms_cosem.protocol.acse.base import AbstractAcseApdu
from dlms_cosem.protocol.acse.user_information import UserInformation

# TODO: It might be a better approach to give the callable and not the class to make a
#   object from bytes. This means we could jack into the creation if needed
#   and also using partials and other for integers etc.
release_reason_from_bytes = partial(
    enumerations.ReleaseRequestReason.from_bytes, byteorder="big"
)


@attr.s(auto_attribs=True)
class ReleaseRequest(AbstractAcseApdu):
    """
    When closing down an Application Association a ReleaseRequest is sent.

    When using ciphering the userinformation can hold an InitiateRequest.
     Even if there is no dedicated key the `InitiateRequestApdu` should be protected
        as above if there is need to protect the RLRQ


    """

    TAG: ClassVar[int] = 98  # Application 2

    PARSE_TAGS: ClassVar[Dict[int, Tuple[str, Callable]]] = {
        0x80: ("reason", release_reason_from_bytes),  # context specific, constricted 0
        0xBE: (
            "user_information",
            UserInformation.from_bytes,
        ),  # Context specific, constructed 30
    }

    reason: Optional[enumerations.ReleaseRequestReason] = attr.ib(default=None)
    user_information: Optional[UserInformation] = attr.ib(default=None)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        # put it in a bytearray to be able to pop.
        rlrq_data = bytearray(source_bytes)

        rlrq_tag = rlrq_data.pop(0)
        if not rlrq_tag == cls.TAG:
            raise ValueError("Bytes are not an RLRQ APDU. TAg is not int(98)")

        rlrq_length = rlrq_data.pop(0)

        if not len(rlrq_data) == rlrq_length:
            raise ValueError(
                "The APDU Data lenght does not correspond " "to length byte"
            )

        # Assumes that the protocol-version is 1 and we don't need to decode it

        # Decode the AARQ  data
        object_dict = dict()
        # use the data in tags to go through the bytes and create objects.
        while True:
            object_tag = rlrq_data.pop(0)
            object_desc = ReleaseRequest.PARSE_TAGS.get(object_tag, None)
            if object_desc is None:
                raise ValueError(
                    f"Could not find object with tag {object_tag} "
                    f"in RLRQ definition"
                )

            object_length = rlrq_data.pop(0)
            object_data = bytes(rlrq_data[:object_length])
            rlrq_data = rlrq_data[object_length:]

            object_name = object_desc[0]
            call: Callable = object_desc[1]

            if call is not None:

                object_data = call(object_data)

            object_dict[object_name] = object_data

            if len(rlrq_data) <= 0:
                break

        return cls(**object_dict)

    def to_bytes(self) -> bytes:
        rlrq_data = bytearray()
        # default value of protocol_version is 1. Only decode if other than 1

        if self.reason is not None:
            rlrq_data.extend(BER.encode(0x80, self.reason.value.to_bytes(1, "big")))
        if self.user_information is not None:
            rlrq_data.extend(BER.encode(0xBE, self.user_information.to_bytes()))
        return BER.encode(self.TAG, rlrq_data)



from typing import *

import attr

from dlms_cosem import security
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.conformance import Conformance


@attr.s(auto_attribs=True)
class InitiateResponse(AbstractXDlmsApdu):
    """
    InitiateResponse ::= SEQUENCE {
        negotiated-quality-of-service [0] IMPLICIT Integer8 OPTIONAL,
        negotiated-dlms-version-number  Unsigned8,
        negotiated-conformance  Conformance, -- Shall be encoded in BER
        server-max-receive-pdu-size   Unsigned16,
        vaa-name  ObjectName

    }
    When using LN referencing the value if vaa-name is always 0x0007
    """

    TAG: ClassVar[int] = 0x08

    negotiated_conformance: Conformance
    server_max_receive_pdu_size: int
    negotiated_dlms_version_number: int = attr.ib(default=6)  # Always 6
    negotiated_quality_of_service: int = attr.ib(default=0)  # not used in dlms.

    @classmethod
    def from_bytes(cls, source_bytes: bytes):

        # Since Initiate response mixes BER and A-XDR we should just "handparse" it.
        if not source_bytes.endswith(b"\x00\x07"):
            raise ValueError("vaa-name in InitateResponse is not \x00\x07")

        data = bytearray(source_bytes[:-2])
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Data is not a InitiateResponse APDU, got apdu tag {tag}")

        use_quality_of_service = data.pop(0)
        if use_quality_of_service:
            quality_of_service = data.pop(0)
        else:
            quality_of_service = 0

        dlms_version = data.pop(0)

        conformance_tag_and_length = data[:3]
        if conformance_tag_and_length != b"\x5f\x1f\x04":
            print(conformance_tag_and_length)
            raise ValueError("Not correct conformance tag and length")

        conformance = Conformance.from_bytes(data[3:-2])

        max_pdu_size = int.from_bytes(data[-2:], "big")

        return cls(
            negotiated_conformance=conformance,
            negotiated_dlms_version_number=dlms_version,
            negotiated_quality_of_service=quality_of_service,
            server_max_receive_pdu_size=max_pdu_size,
        )

    def to_bytes(self) -> bytes:
        # quick and dirty encoding
        out = bytearray()
        out.append(self.negotiated_quality_of_service)
        out.append(self.negotiated_dlms_version_number)
        out.extend(b"\x5f\x1f\x04")
        out.extend(self.negotiated_conformance.to_bytes())
        out.extend(self.server_max_receive_pdu_size.to_bytes(2, "big"))

        return b"\x08" + bytes(out) + b"\x00\x07"


@attr.s(auto_attribs=True)
class GlobalCipherInitiateResponse(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 40

    security_control: security.SecurityControlField
    invocation_counter: int
    ciphered_text: bytes

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but got {tag}")

        length = data.pop(0)
        if length != len(data):
            raise ValueError(f"Octetstring is not of correct length")

        security_control = security.SecurityControlField.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        invocation_counter = int.from_bytes(data[:4], "big")
        ciphered_text = bytes(data[4:])

        return cls(security_control, invocation_counter, ciphered_text)

    def to_bytes(self):
        out = bytearray()
        out.append(self.TAG)

        octet_string_data = bytearray()
        octet_string_data.extend(self.security_control.to_bytes())
        octet_string_data.extend(self.invocation_counter.to_bytes(4, "big"))
        octet_string_data.extend(self.ciphered_text)
        out.append(len(octet_string_data))
        out.extend(octet_string_data)
        return bytes(out)



from dlms_cosem.protocol.xdlms.action import (
    ActionRequestFactory,
    ActionRequestNormal,
    ActionResponseFactory,
    ActionResponseNormal,
    ActionResponseNormalWithData,
    ActionResponseNormalWithError,
)
from dlms_cosem.protocol.xdlms.confirmed_service_error import ConfirmedServiceError
from dlms_cosem.protocol.xdlms.conformance import Conformance
from dlms_cosem.protocol.xdlms.data_notification import DataNotification
from dlms_cosem.protocol.xdlms.exception_response import ExceptionResponse
from dlms_cosem.protocol.xdlms.general_global_cipher import GeneralGlobalCipher
from dlms_cosem.protocol.xdlms.get import (
    GetRequestFactory,
    GetRequestNext,
    GetRequestNormal,
    GetRequestWithList,
    GetResponseFactory,
    GetResponseLastBlock,
    GetResponseLastBlockWithError,
    GetResponseNormal,
    GetResponseNormalWithError,
    GetResponseWithBlock,
    GetResponseWithList,
)
from dlms_cosem.protocol.xdlms.initiate_request import (
    GlobalCipherInitiateRequest,
    InitiateRequest,
)
from dlms_cosem.protocol.xdlms.initiate_response import (
    GlobalCipherInitiateResponse,
    InitiateResponse,
)
from dlms_cosem.protocol.xdlms.invoke_id_and_priority import InvokeIdAndPriority
from dlms_cosem.protocol.xdlms.set import (
    SetRequestFactory,
    SetRequestNormal,
    SetResponseFactory,
    SetResponseNormal,
)

__all__ = [
    "InitiateRequest",
    "DataNotification",
    "GeneralGlobalCipher",
    "InitiateResponse",
    "ConfirmedServiceError",
    "Conformance",
    "GetRequestNormal",
    "GetRequestWithList",
    "GetRequestNext",
    "GetResponseNormal",
    "GetResponseNormalWithError",
    "GetResponseWithBlock",
    "GetResponseLastBlock",
    "GetResponseLastBlockWithError",
    "GetResponseWithList",
    "GetRequestFactory",
    "GetResponseFactory",
    "SetResponseNormal",
    "SetResponseFactory",
    "SetRequestNormal",
    "SetRequestFactory",
    "ExceptionResponse",
    "GlobalCipherInitiateRequest",
    "GlobalCipherInitiateResponse",
    "ActionResponseNormal",
    "ActionResponseNormalWithData",
    "ActionResponseNormalWithError",
    "ActionResponseFactory",
    "ActionRequestNormal",
    "ActionRequestFactory",
    "InvokeIdAndPriority",
]



from typing import *

import attr

from dlms_cosem.hdlc import validators


@attr.s(auto_attribs=True)
class HdlcAddress:
    """
    A client address shall always be expressed on one byte.
    To enable addressing more than one logical device within a single physical device
    and to support the multi-drop configuration the server address may be divided in
    two parts– may be divided into two parts:
    The logical address to address a logical device (separate addressable entity
    within a physical device) makes up the upper HDLC address
    The logical address must always be present.
    The physical address is used to address a physical device ( a physical device on
    a multi-drop)
    The physical address can be omitted it not used.
    """

    logical_address: int = attr.ib(validator=[validators.validate_hdlc_address])
    physical_address: Optional[int] = attr.ib(
        default=None, validator=[validators.validate_hdlc_address]
    )
    address_type: str = attr.ib(
        default="client", validator=[validators.validate_hdlc_address_type]
    )
    extended_addressing: bool = attr.ib(default=False)

    @property
    def length(self):
        """
        The number of bytes the address makes up.
        :return:
        """
        return len(self.to_bytes())

    def to_bytes(self):
        out: List[Optional[int]] = list()
        if self.address_type == "client":
            # shift left 1 bit and set the lsb to mark end of address.
            out.append(((self.logical_address << 1) | 0b00000001))
        else:
            # server address type

            logical_higher, logical_lower = self._split_address(self.logical_address)

            if self.physical_address:
                physical_higher, physical_lower = self._split_address(
                    self.physical_address
                )
                # mark physical lower as end
                physical_lower = physical_lower | 0b00000001
                out.extend(
                    [logical_higher, logical_lower, physical_higher, physical_lower]
                )
            else:
                # no physical address so mark the logial as end.
                logical_lower = logical_lower | 0b00000001
                out.extend([logical_higher, logical_lower])

        out_bytes = list()
        for address in out:
            if address == 0:
                if self.extended_addressing:
                    out_bytes.append(address.to_bytes(1, "big"))
                else:
                    pass
            else:
                out_bytes.append(address.to_bytes(1, "big"))

        return b"".join(out_bytes)

    @staticmethod
    def _split_address(address: int) -> Tuple[Optional[int], int]:
        higher: Optional[int]
        lower: int

        if address > 0b01111111:
            lower = (address & 0b0000000001111111) << 1
            higher = (address & 0b0011111110000000) >> 6

        else:
            lower = address << 1
            higher = 0

        return higher, lower

    @staticmethod
    def _address_to_byte(address: int) -> bytes:
        return address.to_bytes(1, "big")

    @classmethod
    def destination_from_bytes(cls, frame_bytes: bytes, address_type: str):
        destination_address_data, _ = HdlcAddress.find_address_in_frame_bytes(
            frame_bytes
        )
        (
            destination_logical,
            destination_physical,
            destination_length,
        ) = destination_address_data

        return cls(destination_logical, destination_physical, address_type)

    @classmethod
    def source_from_bytes(cls, frame_bytes: bytes, address_type: str):
        _, source_address_data = HdlcAddress.find_address_in_frame_bytes(frame_bytes)

        source_logical, source_physical, source_length = source_address_data
        extended_address = source_length == 4
        return cls(source_logical, source_physical, address_type, extended_address)

    @staticmethod
    def extract_address_bytes(in_data: bytes) -> Tuple[bytes, bytes]:

        address = bytearray()
        data = bytearray(in_data)
        found_whole_address = False
        while not found_whole_address:
            _byte = data.pop(0)
            if bool(_byte & 0b00000001):
                address.append(_byte)
                found_whole_address = True
            else:
                address.append(_byte)
            if len(address) > 4:
                raise ValueError(
                    "Recovered and HDLC address of length longer than 4 bytes"
                )

        return bytes(address), bytes(data)

    @staticmethod
    def find_address_in_frame_bytes(
        hdlc_frame_bytes: bytes,
    ) -> Tuple[Tuple[int, Optional[int], int], Tuple[int, Optional[int], int]]:
        """
        address can be 1, 2 or 4 bytes long. the end byte is indicated by the of
        the last byte LSB being 1
        The first address is the destination address and the seconds is the
        source address.
        :param frame_bytes:
        :return:
        """

        # Find destination address.
        destination_length: int = 1
        destination_logical: int = 0
        destination_physical: Optional[int] = 0
        destination_positions_list: List[Tuple[int, int]] = [(3, 1), (4, 2), (6, 4)]
        address_bytes: bytes
        for pos, _length in destination_positions_list:
            end_byte = hdlc_frame_bytes[pos]
            if bool(end_byte & 0b00000001):
                # Found end byte:
                destination_length = _length
                break
            continue
        if destination_length == 1:
            destinantion_address_bytes = hdlc_frame_bytes[3].to_bytes(1, "big")
            destination_logical = destinantion_address_bytes[0] >> 1
            destination_physical = None

        elif destination_length == 2:
            destinantion_address_bytes = hdlc_frame_bytes[3:5]
            destination_logical = destinantion_address_bytes[0] >> 1
            destination_physical = destinantion_address_bytes[1] >> 1

        elif destination_length == 4:
            destinantion_address_bytes = hdlc_frame_bytes[3:7]
            destination_logical = HdlcAddress.parse_two_byte_address(
                destinantion_address_bytes[:2]
            )
            destination_physical = HdlcAddress.parse_two_byte_address(
                destinantion_address_bytes[2:]
            )

        # Find source address
        source_length: int = 1
        source_logical: int = 0
        source_physical: Optional[int] = 0
        source_position_list: List[Tuple[int, int]] = [
            (item[0] + destination_length, item[1])
            for item in destination_positions_list
        ]
        for pos, _length in source_position_list:
            end_byte = hdlc_frame_bytes[pos]
            if bool(end_byte & 0b00000001):
                # Found end byte:
                source_length = _length
                break
            continue
        if source_length == 1:
            source_address_bytes = hdlc_frame_bytes[3 + destination_length].to_bytes(
                1, "big"
            )
            source_logical = source_address_bytes[0] >> 1
            source_physical = None

        elif source_length == 2:
            source_address_bytes = hdlc_frame_bytes[
                3 + destination_length : 3 + destination_length + source_length
            ]
            source_logical = source_address_bytes[0] >> 1
            source_physical = source_address_bytes[1] >> 1

        elif source_length == 4:
            source_address_bytes = hdlc_frame_bytes[
                3 + destination_length : 3 + destination_length + source_length
            ]
            source_logical = HdlcAddress.parse_two_byte_address(
                source_address_bytes[:2]
            )
            source_physical = HdlcAddress.parse_two_byte_address(
                source_address_bytes[2:]
            )

        return (
            (destination_logical, destination_physical, destination_length),
            (source_logical, source_physical, source_length),
        )

    @staticmethod
    def parse_two_byte_address(address_bytes: bytes):
        if len(address_bytes) != 2:
            raise ValueError(f"Can only parse 2 bytes for address")
        upper = address_bytes[0] >> 1
        lower = address_bytes[1] >> 1

        return lower + (upper << 7)



from dlms_cosem.cosem.attribute_with_selection import CosemAttributeWithSelection
from dlms_cosem.cosem.base import CosemAttribute, CosemMethod
from dlms_cosem.cosem.obis import Obis

__all__ = ["CosemAttribute", "CosemMethod", "Obis", "CosemAttributeWithSelection"]



from dlms_cosem import cosem, enumerations
from dlms_cosem.parsers import ProfileGenericBufferParser


def test_parse_buffer():
    data = b"\x01\x04\x02\x04\t\x0c\x07\xe3\x0c\x1f\x02\x17\x00\x00\x00\xff\xc4\x00\x11\x06\x06\x00\x00\x05\xed\x06\x00\x00\x06T\x02\x04\x00\x11\x06\x06\x00\x00\x05\xed\x06\x00\x00\x06T\x02\x04\x00\x11\x06\x06\x00\x00\x05\xed\x06\x00\x00\x06T\x02\x04\x00\x11\x06\x06\x00\x00\x05\xed\x06\x00\x00\x06T"

    parser = ProfileGenericBufferParser(
        capture_objects=[
            cosem.CosemAttribute(
                interface=enumerations.CosemInterface.CLOCK,
                instance=cosem.Obis(0, 0, 1, 0, 0, 255),
                attribute=2,
            ),
            cosem.CosemAttribute(
                interface=enumerations.CosemInterface.DATA,
                instance=cosem.Obis(0, 0, 96, 10, 1, 255),
                attribute=2,
            ),
            cosem.CosemAttribute(
                interface=enumerations.CosemInterface.REGISTER,
                instance=cosem.Obis(1, 0, 1, 8, 0, 255),
                attribute=2,
            ),
            cosem.CosemAttribute(
                interface=enumerations.CosemInterface.REGISTER,
                instance=cosem.Obis(1, 0, 2, 8, 0, 255),
                attribute=2,
            ),
        ],
        capture_period=60,
    )
    result = parser.parse_bytes(data)

    assert len(result) == 4
    assert len(result[0]) == 4
    assert result[0][0].attribute.attribute == 2
    assert result[0][0].attribute.interface == enumerations.CosemInterface.CLOCK
    assert result[0][0].attribute.instance.to_string() == "0-0:1.0.0.255"
    assert (result[1][0].value - result[0][0].value).total_seconds() == 60 * 60



from datetime import datetime
from typing import *

import attr

from dlms_cosem import cosem, dlms_data, enumerations, time, utils
from dlms_cosem.cosem.capture_object import CaptureObject


@attr.s(auto_attribs=True)
class RangeDescriptor:
    """
    The range descriptor can be used to read buffers of Profile Generic.
    Only buffer element that corresponds to the descriptor shall be returned in a get
    request.


    """

    ACCESS_DESCRIPTOR: ClassVar[int] = 1

    restricting_object: CaptureObject = attr.ib(
        validator=attr.validators.instance_of(CaptureObject)
    )
    from_value: datetime = attr.ib(validator=attr.validators.instance_of(datetime))
    to_value: datetime = attr.ib(validator=attr.validators.instance_of(datetime))
    selected_values: Optional[List[CaptureObject]] = attr.ib(default=None)

    @classmethod
    def from_bytes(cls, source_bytes: bytes) -> "RangeDescriptor":
        data = bytearray(source_bytes)
        access_descriptor = data.pop(0)
        if access_descriptor is not cls.ACCESS_DESCRIPTOR:
            raise ValueError(
                f"Access descriptor {access_descriptor} is not valid for "
                f"RangeDescriptor. It should be {cls.ACCESS_DESCRIPTOR}"
            )
        parsed_data = utils.parse_as_dlms_data(data)

        restricting_object_data = parsed_data[0]
        from_value_data = parsed_data[1]
        to_value_data = parsed_data[2]
        selected_values_data = parsed_data[3]

        restricting_cosem_attribute = cosem.CosemAttribute(
            interface=enumerations.CosemInterface(restricting_object_data[0]),
            instance=cosem.Obis.from_bytes(restricting_object_data[1]),
            attribute=restricting_object_data[2],
        )
        restricting_object = CaptureObject(
            cosem_attribute=restricting_cosem_attribute,
            data_index=restricting_object_data[3],
        )
        from_dt, clock_status = time.datetime_from_bytes(from_value_data)
        to_dt, clock_status = time.datetime_from_bytes(to_value_data)
        if selected_values_data:
            raise NotImplementedError()
        else:
            selected_values = None

        return cls(
            restricting_object=restricting_object,
            from_value=from_dt,
            to_value=to_dt,
            selected_values=selected_values,
        )

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.ACCESS_DESCRIPTOR)
        out.extend(b"\x02\x04")  # structure of 4 elements
        out.extend(self.restricting_object.to_bytes())
        out.extend(
            dlms_data.OctetStringData(
                time.datetime_to_bytes(self.from_value)
            ).to_bytes()
        )
        out.extend(
            dlms_data.OctetStringData(time.datetime_to_bytes(self.to_value)).to_bytes()
        )
        if not self.selected_values:
            out.extend(b"\x01\x00")  # empty array for selected values means all columns
        else:
            raise NotImplementedError()
            # TODO: implement selected values

        return bytes(out)


def validate_unsigned_double_long_int(instance, attribute, value):
    if 0 >= value >= 0xFFFFFFFF:
        raise ValueError(
            f"{value} is not withing the limits of a unsigned double long integer"
        )


def validate_unsigned_long_int(instance, attribute, value):
    if 0 >= value >= 0xFFFF:
        raise ValueError(
            f"{value} is not withing the limits of a unsigned long integer"
        )


@attr.s(auto_attribs=True)
class EntryDescriptor:
    """
    The entry descriptor limits response data by entries.
    It is possible to limit the entries and also the columns returned.
    The from/to_selected_value limits the columns returned from/to_entry limits the
    entries.

    Numbering of selected values and entries start from 1.
    Setting to_entry=0 or to_selected_value=0 requests the highest possible value.
    """

    ACCESS_DESCRIPTOR: ClassVar[int] = 2

    from_entry: int = attr.ib(
        validator=[validate_unsigned_double_long_int, attr.validators.instance_of(int)]
    )
    to_entry: int = attr.ib(
        validator=[validate_unsigned_double_long_int, attr.validators.instance_of(int)],
        default=0,
    )
    from_selected_value: int = attr.ib(
        validator=[validate_unsigned_long_int, attr.validators.instance_of(int)],
        default=1,
    )
    to_selected_value: int = attr.ib(
        validator=[validate_unsigned_long_int, attr.validators.instance_of(int)],
        default=0,
    )

    @classmethod
    def from_bytes(cls, source_bytes) -> "EntryDescriptor":
        raise NotImplementedError()

    def to_bytes(self) -> bytes:
        raise NotImplementedError()


@attr.s(auto_attribs=True)
class AccessDescriptorFactory:

    """
    Handles the selection of parsing the first byte to find what kind of access
    descriptor it is and returns the object.
    """

    @staticmethod
    def from_bytes(source_bytes: bytes) -> Union[RangeDescriptor, EntryDescriptor]:

        access_descriptor = source_bytes[0]
        if access_descriptor == 1:
            return RangeDescriptor.from_bytes(source_bytes)
        elif access_descriptor == 2:
            return EntryDescriptor.from_bytes(source_bytes)
        else:
            raise ValueError(f"{access_descriptor} is not a valid access descriptor")



def validate_information_sequence_number(instance, attribute, value):
    if not 0 <= value <= 7:
        raise ValueError(f"Sequence number can only be between 0-7. Got {value}")


def validate_hdlc_address_type(instance, attribute, value):
    if value not in ["client", "server"]:
        raise ValueError("HdlcAddress type can only be client or server.")


def validate_hdlc_address(instance, attribute, value):
    """
    Client addresses should always be expressed in 1 byte.
    With the marking bit that leaves 7 bits for address.

    A server address can be expressed in 1 or 2 bytes (well technically 2 or 4 but that
    is including both the logical and physical address. Each value is limited to max 2 bytes
    but 7 bits in each byte.


    """
    if (attribute.name == "physical_address") & (value is None):
        # we allow physical address to be none.
        return

    if instance.address_type == "client":
        address_limit = 0b01111111

    else:  # server
        address_limit = 0b0011111111111111

    if value > address_limit:
        raise ValueError(
            f"Hdlc {instance.address_type} address cannot be higher "
            f"than {address_limit}, but is {value}"
        )

    if value < 0:
        raise ValueError("Hdlc address cannot have a negative value.")



import pytest

from dlms_cosem import cosem, enumerations
from dlms_cosem.protocol import xdlms


class TestActionRequestNormal:
    def test_transform_bytes(self):
        data = b'\xc3\x01\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x01\t\x11\x10\x00\x00\x1a\x90\xe6\xd2"\x1f\xa2\xfd\x85\xee\xd6\x1a\xcc"'
        action = xdlms.ActionRequestNormal(
            cosem_method=cosem.CosemMethod(
                interface=enumerations.CosemInterface.ASSOCIATION_LN,
                instance=cosem.Obis(a=0, b=0, c=40, d=0, e=0, f=255),
                method=1,
            ),
            data=b'\t\x11\x10\x00\x00\x1a\x90\xe6\xd2"\x1f\xa2\xfd\x85\xee\xd6\x1a\xcc"',
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=0, confirmed=True, high_priority=True
            ),
        )

        assert data == action.to_bytes()
        assert action == xdlms.ActionRequestNormal.from_bytes(data)

    def test_transform_bytes_without_data(self):
        data = b"\xc3\x01\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x00"
        action = xdlms.ActionRequestNormal(
            cosem_method=cosem.CosemMethod(
                interface=enumerations.CosemInterface.ASSOCIATION_LN,
                instance=cosem.Obis(a=0, b=0, c=40, d=0, e=0, f=255),
                method=1,
            ),
            data=None,
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=0, confirmed=True, high_priority=True
            ),
        )

        assert data == action.to_bytes()
        assert action == xdlms.ActionRequestNormal.from_bytes(data)

    def test_wrong_tag_raises_valueerror(self):
        data = b'\xc4\x01\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x01\t\x11\x10\x00\x00\x1a\x90\xe6\xd2"\x1f\xa2\xfd\x85\xee\xd6\x1a\xcc"'

        with pytest.raises(ValueError):
            xdlms.ActionRequestNormal.from_bytes(data)

    def test_wrong_request_type_raises_valueerror(self):
        data = b'\xc3\x02\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x01\t\x11\x10\x00\x00\x1a\x90\xe6\xd2"\x1f\xa2\xfd\x85\xee\xd6\x1a\xcc"'

        with pytest.raises(ValueError):
            xdlms.ActionRequestNormal.from_bytes(data)


class TestActionRequestFactory:
    def test_normal_with_data(self):
        data = b'\xc3\x01\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x01\t\x11\x10\x00\x00\x1a\x90\xe6\xd2"\x1f\xa2\xfd\x85\xee\xd6\x1a\xcc"'
        action = xdlms.ActionRequestFactory.from_bytes(data)
        assert isinstance(action, xdlms.ActionRequestNormal)

    def test_normal_without_data(self):
        data = b"\xc3\x01\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x00"
        action = xdlms.ActionRequestFactory.from_bytes(data)
        assert isinstance(action, xdlms.ActionRequestNormal)
        assert action.data is None

    def test_wrong_tag_raises_valueerror(self):
        data = b"\xc4\x01\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x00"
        with pytest.raises(ValueError):
            xdlms.ActionRequestFactory.from_bytes(data)

    def test_any_other_type_than_normal_raises_notimplementederror(self):
        data = b"\xc3\x02\xc0\x00\x0f\x00\x00(\x00\x00\xff\x01\x00"
        with pytest.raises(NotImplementedError):
            xdlms.ActionRequestFactory.from_bytes(data)


class TestActionResponseNormal:
    def test_transform_bytes(self):
        data = b"\xc7\x01\xc0\x00\x00"
        action = xdlms.ActionResponseNormal(
            enumerations.ActionResultStatus.SUCCESS,
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=0, confirmed=True, high_priority=True
            ),
        )
        assert data == action.to_bytes()
        assert action == xdlms.ActionResponseNormal.from_bytes(data)

    def test_wrong_tag_raises_valueerror(self):
        data = b"\xc8\x01\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormal.from_bytes(data)

    def test_wrong_type_raises_valueerror(self):
        data = b"\xc7\x02\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormal.from_bytes(data)

    def test_has_data_raises_valueerror(self):
        data = b"\xc7\x01\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormal.from_bytes(data)


class TestActionResponseNormalWithData:
    def test_transform_bytes(self):
        data = b"\xc7\x01\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        action = xdlms.ActionResponseNormalWithData(
            enumerations.ActionResultStatus.SUCCESS,
            data=bytearray(
                b"\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
            ),
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=0, confirmed=True, high_priority=True
            ),
        )
        assert data == action.to_bytes()
        assert action == xdlms.ActionResponseNormalWithData.from_bytes(data)

    def test_wrong_tag_raises_valueerror(self):
        data = b"\xc8\x01\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithData.from_bytes(data)

    def test_wrong_type_raises_valueerror(self):
        data = b"\xc7\x02\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithData.from_bytes(data)

    def test_no_data_raises_valueerror(self):
        data = b"\xc7\x01\xc0\x00\x00"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithData.from_bytes(data)

    def test_holds_error_instead_of_data_raises_valueerror(self):
        data = b"\xc7\x01\xc0\x00\x01\x01\x01"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithData.from_bytes(data)


class TestActionResponseWithError:
    def test_transform_bytes(self):
        data = b"\xc7\x01\xc0\x00\x01\x01\xfa"
        action = xdlms.ActionResponseNormalWithError(
            enumerations.ActionResultStatus.SUCCESS,
            error=enumerations.DataAccessResult.OTHER_REASON,
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=0, confirmed=True, high_priority=True
            ),
        )
        assert data == action.to_bytes()
        assert action == xdlms.ActionResponseNormalWithError.from_bytes(data)

    def test_wrong_tag_raises_valueerror(self):
        data = b"\xc8\x01\xc0\x00\x01\x01\xfa"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithError.from_bytes(data)

    def test_wrong_type_raises_valueerror(self):
        data = b"\xc7\x02\xc0\x00\x01\x01\xfa"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithError.from_bytes(data)

    def test_no_data_raises_valueerror(self):
        data = b"\xc7\x01\xc0\x00\x00"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithError.from_bytes(data)

    def test_holds_data_instead_of_error_raises_valueerror(self):
        data = b"\xc7\x01\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        with pytest.raises(ValueError):
            xdlms.ActionResponseNormalWithError.from_bytes(data)


class TestActionResponseFactory:
    def test_parse_action_response_normal(self):
        data = b"\xc7\x01\xc0\x00\x00"
        action = xdlms.ActionResponseFactory.from_bytes(data)
        assert isinstance(action, xdlms.ActionResponseNormal)

    def test_parse_action_response_normal_with_data(self):
        data = b"\xc7\x01\xc0\x00\x01\x00\t\x11\x10\x00\x00\x1a\xfd\xe8\x85{r\x8a4\x99\x10j\xa6e\xd1"
        action = xdlms.ActionResponseFactory.from_bytes(data)
        assert isinstance(action, xdlms.ActionResponseNormalWithData)

    def test_parse_action_response_normal_with_error(self):
        data = b"\xc7\x01\xc0\x00\x01\x01\xfa"
        action = xdlms.ActionResponseFactory.from_bytes(data)
        assert isinstance(action, xdlms.ActionResponseNormalWithError)

    def test_wrong_tag_raises_valueerror(self):
        data = b"\xc8\x01\xc0\x00\x01\x01\xfa"
        with pytest.raises(ValueError):
            xdlms.ActionResponseFactory.from_bytes(data)

    def test_type_other_than_normal_raises_not_implemented_error(self):
        data = b"\xc7\x02\xc0\x00\x01\x01\xfa"
        with pytest.raises(NotImplementedError):
            xdlms.ActionResponseFactory.from_bytes(data)



import abc


class AbstractXDlmsApdu(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def from_bytes(cls, source_bytes: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    def to_bytes(self) -> bytes:
        raise NotImplementedError()



from typing import *

from dlms_cosem import a_xdr, dlms_data
from dlms_cosem.dlms_data import decode_variable_integer


def parse_as_dlms_data(data: bytes):
    data_decoder = a_xdr.AXdrDecoder(
        encoding_conf=a_xdr.EncodingConf(
            attributes=[a_xdr.Sequence(attribute_name="data")]
        )
    )
    return data_decoder.decode(data)["data"]


def parse_dlms_object(source_bytes: bytes) -> List[int]:
    """
    Some DLMS object attributes contain data structures. The items are not self
    descriptive and we cannot use the normal parser since we dont know what items to
    parse as the data tag is not included.

    But is seems they are always integers so we can parse them as a list of integers.
    """
    values = list()
    data = bytearray(source_bytes)
    tag = data.pop(0)
    allowed_dlms_object_tags = [dlms_data.DataArray.TAG, dlms_data.DataStructure.TAG]
    if tag not in allowed_dlms_object_tags:
        raise ValueError(
            f"You cannot use the dlms object parse "
            f"with {dlms_data.DlmsDataFactory.get_data_class(tag)}"
        )
    length, rest = decode_variable_integer(data)
    data = rest
    for i in range(0, length):
        values.append(data.pop(0))

    return values



import pytest

from dlms_cosem import enumerations
from dlms_cosem.protocol import xdlms
from dlms_cosem.protocol.acse import ReleaseResponse


class TestDecodeRLRE:
    def test_simple(self):
        data = b"c\x03\x80\x01\x00"
        rlre = ReleaseResponse.from_bytes(data)
        assert rlre.reason == enumerations.ReleaseResponseReason.NORMAL
        assert rlre.user_information is None

    def test_with_initiate_response(self):
        data = b"c\x16\x80\x01\x00\xbe\x11\x04\x0f\x08\x01\x00\x06_\x1f\x04\x00\x00\x1e\x1d\x04\xc8\x00\x07"
        rlre = ReleaseResponse.from_bytes(data)
        assert rlre.reason == enumerations.ReleaseResponseReason.NORMAL
        assert isinstance(rlre.user_information.content, xdlms.InitiateResponse)

    def test_with_ciphered_initiate_response(self):
        data = bytes.fromhex(
            "6328800100BE230421281F3001234567891214A0845E475714383F65BC19745CA235906525E4F3E1C893"
        )
        rlre = ReleaseResponse.from_bytes(data)
        assert rlre.reason == enumerations.ReleaseResponseReason.NORMAL
        assert isinstance(
            rlre.user_information.content, xdlms.GlobalCipherInitiateResponse
        )



import pytest

from dlms_cosem import dlms_data, enumerations, exceptions, security, state
from dlms_cosem.security import NoSecurityAuthentication
from dlms_cosem.connection import (
    DlmsConnection,
    XDlmsApduFactory,
    make_client_to_server_challenge, DlmsConnectionSettings,
)
from dlms_cosem.exceptions import LocalDlmsProtocolError
from dlms_cosem.protocol import acse, xdlms
from dlms_cosem.protocol.xdlms import Conformance


def test_conformance_exists_on_simple_init():
    c = DlmsConnection(
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )
    assert c.conformance is not None
    assert not c.conformance.general_protection
    assert c.state.current_state == state.NO_ASSOCIATION


def test_conformance_protection_is_set_when_passing_encryption_key():
    c = DlmsConnection(
        global_encryption_key=b"1234",
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )
    assert c.conformance.general_protection
    assert c.state.current_state == state.NO_ASSOCIATION


def test_negotiated_conformance_is_updated():
    c = DlmsConnection(
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )
    c.send(c.get_aarq())
    c.receive_data(
        acse.ApplicationAssociationResponse(
            result=enumerations.AssociationResult.ACCEPTED,
            result_source_diagnostics=enumerations.AcseServiceUserDiagnostics.NULL,
            user_information=acse.UserInformation(
                content=xdlms.InitiateResponse(
                    negotiated_conformance=Conformance(
                        general_protection=True, general_block_transfer=True
                    ),
                    server_max_receive_pdu_size=500,
                )
            ),
        ).to_bytes()
    )
    c.next_event()
    assert c.conformance.general_protection
    assert c.conformance.general_block_transfer
    assert c.max_pdu_size == 500
    assert c.state.current_state == state.READY


def test_settings_exists_on_simple_init():
    c = DlmsConnection(
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )
    assert c.settings is not None


def test_settings_empty_system_title_in_general_glo_cipher_false(get_request: xdlms.GetRequestNormal):
    """
    Make sure that system_title is is used when protecting APDUs with default connection settings.
    """
    system_title = b"12345678"
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=system_title,
        authentication=NoSecurityAuthentication(),
        global_encryption_key=b"1111111111111111",
        global_authentication_key=b"0000000000000000"
    )

    assert c.settings is not None

    print(c.settings)

    ciphered = c.protect(get_request)
    print(ciphered)
    assert ciphered.system_title is not None
    assert ciphered.system_title == system_title


def test_settings_empty_system_title_in_general_glo_cipher_true(get_request: xdlms.GetRequestNormal):
    """
    Make sure that system_title is not used when protecting APDUs if the connections settings is to leave it empty
    """
    system_title = b"12345678"
    settings = DlmsConnectionSettings(empty_system_title_in_general_glo_ciphering=True)

    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=system_title,
        authentication=NoSecurityAuthentication(),
        global_encryption_key=b"1111111111111111",
        global_authentication_key=b"0000000000000000",
        settings=settings,
    )

    assert c.settings is not None

    ciphered = c.protect(get_request)
    assert ciphered.system_title is None


def test_cannot_re_associate(aarq: acse.ApplicationAssociationRequest):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    with pytest.raises(LocalDlmsProtocolError):
        c.send(aarq)


def test_can_release_in_ready_state(rlrq: acse.ReleaseRequest):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.send(rlrq)
    assert c.state.current_state == state.AWAITING_RELEASE_RESPONSE


def test_receive_rlre_terminates_association(rlre: acse.ReleaseResponse):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.AWAITING_RELEASE_RESPONSE),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )
    c.receive_data(rlre.to_bytes())
    c.next_event()
    assert c.state.current_state == state.NO_ASSOCIATION


def test_can_send_get_when_ready(get_request: xdlms.GetRequestNormal):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.send(get_request)
    assert c.state.current_state == state.AWAITING_GET_RESPONSE


def test_receive_get_response_sets_state_to_ready():
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.AWAITING_GET_RESPONSE),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )
    c.receive_data(b"\xc4\x01\xc1\x00\x06\x00\x00\x13\x91")
    c.next_event()
    assert c.state.current_state == state.READY


def test_set_request_sets_state_in_waiting_for_set_response(
        set_request: xdlms.SetRequestNormal,
):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.send(set_request)
    assert c.state.current_state == state.AWAITING_SET_RESPONSE


def test_set_response_sets_state_in_ready(set_response: xdlms.SetResponseNormal):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.AWAITING_SET_RESPONSE),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.receive_data(set_response.to_bytes())
    c.next_event()
    assert c.state.current_state == state.READY


def test_can_send_action_request_in_ready(action_request: xdlms.ActionRequestNormal):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.send(action_request)
    assert c.state.current_state == state.AWAITING_ACTION_RESPONSE


def test_action_response_normal_sets_ready_when_awaiting_action_resoponse():
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.AWAITING_ACTION_RESPONSE),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.receive_data(
        xdlms.ActionResponseNormal(
            status=enumerations.ActionResultStatus.SUCCESS,
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=1, confirmed=True, high_priority=True
            ),
        ).to_bytes()
    )
    c.next_event()
    assert c.state.current_state == state.READY


def test_action_response_normal_with_error_sets_ready_when_awaiting_action_resoponse():
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.AWAITING_ACTION_RESPONSE),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.receive_data(
        xdlms.ActionResponseNormalWithError(
            status=enumerations.ActionResultStatus.SUCCESS,
            error=enumerations.DataAccessResult.OTHER_REASON,
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=1, confirmed=True, high_priority=True
            ),
        ).to_bytes()
    )
    c.next_event()
    assert c.state.current_state == state.READY


def test_action_response_normal_with_data_sets_ready_when_awaiting_action_resoponse():
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.AWAITING_ACTION_RESPONSE),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )

    c.receive_data(
        xdlms.ActionResponseNormalWithData(
            status=enumerations.ActionResultStatus.SUCCESS,
            data=b"\x12\x00\x00",
            invoke_id_and_priority=xdlms.InvokeIdAndPriority(
                invoke_id=1, confirmed=True, high_priority=True
            ),
        ).to_bytes()
    )
    c.next_event()
    assert c.state.current_state == state.READY


def test_receive_exception_response_sets_state_to_ready(
        exception_response: xdlms.ExceptionResponse,
):
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.AWAITING_GET_RESPONSE),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
    )
    c.receive_data(exception_response.to_bytes())
    c.next_event()
    assert c.state.current_state == state.READY


def test_hls_is_started_automatically(
        connection_with_hls: DlmsConnection,
        ciphered_hls_aare: acse.ApplicationAssociationResponse,
):
    # Force state into awaiting response
    connection_with_hls.state.current_state = state.AWAITING_ASSOCIATION_RESPONSE
    connection_with_hls.receive_data(ciphered_hls_aare.to_bytes())
    connection_with_hls.next_event()
    assert (
            connection_with_hls.state.current_state
            == state.SHOULD_SEND_HLS_SEVER_CHALLENGE_RESULT
    )


def test_hls_fails(connection_with_hls: DlmsConnection):
    # Force state into awaiting response
    connection_with_hls.state.current_state = state.AWAITING_HLS_CLIENT_CHALLENGE_RESULT
    connection_with_hls.meter_system_title = b"12345678"
    connection_with_hls.meter_invocation_counter = 1
    failing_action_response = xdlms.ActionResponseNormal(
        status=enumerations.ActionResultStatus.OTHER_REASON
    )
    ciphered = security.encrypt(
        security_control=connection_with_hls.security_control,
        system_title=connection_with_hls.meter_system_title,
        auth_key=connection_with_hls.global_authentication_key,
        key=connection_with_hls.global_encryption_key,
        invocation_counter=2,
        plain_text=failing_action_response.to_bytes(),
    )
    ciphered_action_response = xdlms.GeneralGlobalCipher(
        security_control=connection_with_hls.security_control,
        system_title=connection_with_hls.meter_system_title,
        invocation_counter=2,
        ciphered_text=ciphered,
    )
    connection_with_hls.receive_data(ciphered_action_response.to_bytes())
    connection_with_hls.next_event()
    assert connection_with_hls.state.current_state == state.NO_ASSOCIATION


def test_rejection_resets_connection_state(
        connection_with_hls: DlmsConnection,
        ciphered_hls_aare: acse.ApplicationAssociationResponse,
):
    connection_with_hls.state.current_state = state.AWAITING_ASSOCIATION_RESPONSE
    ciphered_hls_aare.result = enumerations.AssociationResult.REJECTED_PERMANENT
    connection_with_hls.receive_data(ciphered_hls_aare.to_bytes())
    connection_with_hls.next_event()
    assert connection_with_hls.state.current_state == state.NO_ASSOCIATION


def test_rlrq_raises_norlrqrlreerror_when_settings_use_rlrq_rlre_is_false():
    settings = DlmsConnectionSettings(use_rlrq_rlre=False)
    c = DlmsConnection(
        state=state.DlmsConnectionState(current_state=state.READY),
        client_system_title=b"12345678",
        authentication=NoSecurityAuthentication(),
        settings=settings
    )
    rlrq = c.get_rlrq()
    with pytest.raises(exceptions.NoRlrqRlreError):
        c.send(rlrq)

    assert c.state.current_state == state.NO_ASSOCIATION


class TestPreEstablishedAssociation:
    def test_state_is_ready_in_init(self):
        c = DlmsConnection.with_pre_established_association(
            conformance=Conformance(
                priority_management_supported=True,
                attribute_0_supported_with_get=True,
                block_transfer_with_action=True,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=True,
                multiple_references=True,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            )
        )

        assert c.state.current_state == state.READY

    def test_not_able_to_send_aarq(self, aarq: acse.ApplicationAssociationRequest):
        c = DlmsConnection.with_pre_established_association(conformance=Conformance())

        with pytest.raises(exceptions.PreEstablishedAssociationError):
            c.send(aarq)

    def test_not_able_to_send_rlrq(self, rlrq: acse.ReleaseRequest):
        c = DlmsConnection.with_pre_established_association(conformance=Conformance())

        with pytest.raises(exceptions.PreEstablishedAssociationError):
            c.send(rlrq)


class TestXDlmsApduFactory:
    def test_nonexistent_tag_raises_key_error(self):
        dumb_data = bytearray([255, 1, 10, 10, 23])
        with pytest.raises(KeyError):
            XDlmsApduFactory.apdu_from_bytes(dumb_data)


class TestMakeClientToServerChallenge:
    def test_standard_length(self):
        challenge = make_client_to_server_challenge()
        assert len(challenge) == 8

    def test_hls_gmac_returns_correct_bytes(self):
        challenge = make_client_to_server_challenge(16)
        assert challenge
        assert len(challenge) == 16
        assert type(challenge) == bytes

    def test_too_short_length_raises_value_error(self):
        with pytest.raises(ValueError):
            make_client_to_server_challenge(7)

    def test_too_long_length_raises_value_error(self):
        with pytest.raises(ValueError):
            make_client_to_server_challenge(65)



from typing import *


class BER:

    """
    BER encoding consists of a TAG ID, Length and data
    Simple implementation that calculates the lenght.
    """

    @staticmethod
    def encode(tag: Union[int, bytes], data: Union[bytearray, bytes]):

        if isinstance(tag, int):
            # Simplification since we now just use ints as tags when they are single
            # bytes.
            _tag_bytes = tag.to_bytes(1, "big")
        else:
            _tag_bytes = tag

        if data is None:
            return b""

        if not isinstance(data, (bytes, bytearray)):
            raise ValueError(
                f"BER encoding requires bytes or bytearray, got {data!r} of {type(data)}"
            )

        length = len(data).to_bytes(1, "big")
        if length == 0:
            return b""

        return b"".join([_tag_bytes, length, data])

    @staticmethod
    def decode(_bytes: bytes, tag_length: int = 1) -> Tuple[bytes, int, bytes]:
        input = bytearray(_bytes)
        tag = b"".join([input.pop(0).to_bytes(1, "big") for _ in range(tag_length)])
        length = input.pop(0)
        data = input
        if len(data) != length:
            raise ValueError(
                f"BER-decoding failed. Length byte {length} does "
                f"not correspond to length of data {data}"
            )
        return tag, length, data



import socket

import pytest

from dlms_cosem.io import BlockingTcpIO, TcpTransport
from dlms_cosem.exceptions import CommunicationError


class TestBlockingTcpTransport:

    host = "localhost"
    port = 10000
    client_logical_address = 1
    server_logical_address = 1

    def test_can_connect(self):
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind((self.host, self.port))
        server_socket.listen(1)
        io = BlockingTcpIO(host=self.host, port=self.port)
        transport = TcpTransport(
            self.client_logical_address, self.server_logical_address, io
        )
        transport.connect()
        assert transport.io.tcp_socket

    def test_connect_on_connected_raises(self):
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind((self.host, self.port))
        server_socket.listen(1)

        io = BlockingTcpIO(host=self.host, port=self.port)
        transport = TcpTransport(
            self.client_logical_address, self.server_logical_address, io
        )
        transport.connect()
        with pytest.raises(RuntimeError):
            transport.connect()

    def test_cant_connect_raises_communications_error(self):
        io = BlockingTcpIO(host=self.host, port=self.port)
        transport = TcpTransport(
            self.client_logical_address, self.server_logical_address, io
        )
        with pytest.raises(CommunicationError):
            transport.connect()

    def test_disconnect(self):
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind((self.host, self.port))
        server_socket.listen(1)

        io = BlockingTcpIO(host=self.host, port=self.port)
        transport = TcpTransport(
            self.client_logical_address, self.server_logical_address, io
        )
        transport.connect()
        transport.disconnect()
        assert transport.io.tcp_socket is None

    def test_disconnect_is_noop_if_disconnected(self):
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind((self.host, self.port))
        server_socket.listen(1)

        io = BlockingTcpIO(host=self.host, port=self.port)
        transport = TcpTransport(
            self.client_logical_address, self.server_logical_address, io
        )
        transport.connect()
        transport.disconnect()
        transport.disconnect()
        assert transport.io.tcp_socket is None



import datetime
from typing import *

import attr

import dlms_cosem.time as dlmstime
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu


@attr.s(auto_attribs=True)
class LongInvokeIdAndPriority:
    """
    Unsigned 32 bits

     - bit 0-23: Long Invoke ID
     - bit 25-27: Reserved
     - bit 28: Self descriptive -> 0=Not Self Descriptive, 1= Self-descriptive
     - bit 29: Processing options -> 0 = Continue on Error, 1=Break on Error
     - bit 30: Service class -> 0 = Unconfirmed, 1 = Confirmed
     - bit 31 Priority, -> 0 = normal, 1 = high.

    :param int long_invoke_id: Long Invoke ID
    :param bool self_descriptive: Indicates if self descriptive  `DEFAULT=False`
    :param bool confirmed: Indicates if confirmed. `DEFAULT=False`
    :param bool prioritized: Indicates if prioritized. `DEFAULT=False`
    :param bool break_on_error: Indicates id should break in error. `DEFAULT=True`

    """

    long_invoke_id: int
    prioritized: bool = attr.ib(default=False)
    confirmed: bool = attr.ib(default=False)
    self_descriptive: bool = attr.ib(default=False)
    break_on_error: bool = attr.ib(default=False)

    @classmethod
    def from_bytes(cls, bytes_data):
        if len(bytes_data) != 4:
            raise ValueError(
                f"LongInvokeIdAndPriority is 4 bytes long,"
                f" received: {len(bytes_data)}"
            )

        long_invoke_id = int.from_bytes(bytes_data[1:], "big")
        status_byte = bytes_data[0]
        prioritized = bool(status_byte & 0b10000000)
        confirmed = bool(status_byte & 0b01000000)
        break_on_error = bool(status_byte & 0b00100000)
        self_descriptive = bool(status_byte & 0b00010000)

        return cls(
            long_invoke_id=long_invoke_id,
            prioritized=prioritized,
            confirmed=confirmed,
            break_on_error=break_on_error,
            self_descriptive=self_descriptive,
        )

    def to_bytes(self) -> bytes:
        status = 0
        if self.prioritized:
            status = status | 0b10000000
        if self.confirmed:
            status = status | 0b01000000
        if self.break_on_error:
            status = status | 0b00100000
        if self.self_descriptive:
            status = status | 0b00010000
        return status.to_bytes(1, "big") + self.long_invoke_id.to_bytes(3, "big")


@attr.s(auto_attribs=True)
class DataNotification(AbstractXDlmsApdu):
    """
    The DataNotification APDU is used by the DataNotification service.
    It is used to push data from a server (meter) to the client (amr-system).
    It is an unconfirmable service.

    A DataNotification APDU, if to large, can be sent using the general block
    transfer method.

    :param `LongInvokeAndPriority` long_invoke_id_and_priority: The long invoke
        id is a reference to the server invocation. self_descriptive,
        break_on_error and prioritized are not used for Datanotifications.
    :param datetime.datetime date_time: Indicates the time the DataNotification
        was sent. Is optional.
    :param `bytes` body: Push data.
    """

    TAG = 15

    long_invoke_id_and_priority: LongInvokeIdAndPriority
    date_time: Optional[datetime.datetime]
    body: bytes

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Data is not a DataNotification APDU. Expected tag={cls.TAG} but got {tag}"
            )
        long_invoke_id_data = data[:4]
        long_invoke_id = LongInvokeIdAndPriority.from_bytes(bytes(long_invoke_id_data))
        data = data[4:]
        has_datetime = bool(data.pop(0))
        if has_datetime:
            dn_datetime_data = data[:12]
            data = data[12:]
            dn_datetime, _ = dlmstime.datetime_from_bytes(dn_datetime_data)
        else:
            dn_datetime = None
        return cls(
            long_invoke_id_and_priority=long_invoke_id,
            date_time=dn_datetime,
            body=bytes(data),
        )

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.extend(self.long_invoke_id_and_priority.to_bytes())
        if self.date_time:
            out.extend(b"\x01")
            out.extend(dlmstime.datetime_to_bytes(self.date_time))
        else:
            out.extend(b"\x00")
        out.extend(self.body)
        return bytes(out)



from dlms_cosem.connection import DlmsConnectionSettings
from dlms_cosem.security import NoSecurityAuthentication
from dlms_cosem.client import DlmsClient
from dlms_cosem.io import BlockingTcpIO, TcpTransport


class TestDlmsClient:
    def test_client_invocation_counter_property(self):
        transport = TcpTransport(
            io=BlockingTcpIO(host="localhost", port=4059),
            client_logical_address=1,
            server_logical_address=1,
        )
        client = DlmsClient(
            client_initial_invocation_counter=500,
            transport=transport,
            authentication=NoSecurityAuthentication(),
        )

        assert client.client_invocation_counter == 500

    def test_client_invocation_counter_setter(self):
        transport = TcpTransport(
            io=BlockingTcpIO(host="localhost", port=4059),
            client_logical_address=1,
            server_logical_address=1,
        )
        client = DlmsClient(
            client_initial_invocation_counter=500,
            transport=transport,
            authentication=NoSecurityAuthentication(),
        )
        client.client_invocation_counter = 1000
        assert client.client_invocation_counter == 1000
        assert client.dlms_connection.client_invocation_counter == 1000


class TestDlmsClientWithConnectionSettings:

    def test_can_get_settings_from_client(self):
        settings = DlmsConnectionSettings(empty_system_title_in_general_glo_ciphering=True)

        transport = TcpTransport(
            io=BlockingTcpIO(host="localhost", port=4059),
            client_logical_address=1,
            server_logical_address=1,
        )


        client = DlmsClient(
            client_initial_invocation_counter=500,
            transport=transport,
            authentication=NoSecurityAuthentication(),
            connection_settings=settings
        )

        assert client.dlms_connection.settings.empty_system_title_in_general_glo_ciphering == True



from typing import *

import attr


@attr.s(auto_attribs=True)
class InvokeIdAndPriority:
    """
    :parameter invoke_id: It is allowed to send several requests to the server (meter)
        if the lower layers support it, before listening for the response. To be able to
        correlate an answer to a request the invoke_id is used. It is copied in the
        response from the server.

    :parameter confirmed: Indicates if the service is confirmed. Mostly it is.

    :parameter high_priority: When sending several requests to the server (meter) it is
        possible to mark some of them as high priority. These response from the requests
        will be sent back before the ones with normal priority. Handling of priority is
        a negotiable feature in the Conformance block during Application Association.
        If the server (meter) does not support priority it will treat all requests with
        high priority as normal priority.

    """

    invoke_id: int = attr.ib(default=1)
    confirmed: bool = attr.ib(default=True)
    high_priority: bool = attr.ib(default=True)

    LENGTH: ClassVar[int] = 1

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        if len(source_bytes) != cls.LENGTH:
            raise ValueError(
                f"Length of data does not correspond with class LENGTH. "
                f"Should be {cls.LENGTH}, got {len(source_bytes)}"
            )

        val = int.from_bytes(source_bytes, "big")
        invoke_id = val & 0b00001111
        confirmed = bool(val & 0b01000000)
        high_priority = bool(val & 0b10000000)
        return cls(
            invoke_id=invoke_id, confirmed=confirmed, high_priority=high_priority
        )

    def to_bytes(self) -> bytes:
        out = self.invoke_id
        out += self.confirmed << 6
        out += self.high_priority << 7
        return out.to_bytes(1, "big")



from typing import *

import attr

from dlms_cosem import cosem
from dlms_cosem import enumerations as enums
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.invoke_id_and_priority import InvokeIdAndPriority

"""
Set-Request ::= CHOICE
{
set-request-normal                          [1] IMPLICIT Set-Request-Normal,
set-request-with-first-datablock            [2] IMPLICIT Set-Request-With-First-Datablock,
set-request-with-datablock                  [3] IMPLICIT Set-Request-With-Datablock,
set-request-with-list                       [4] IMPLICIT Set-Request-With-List,
set-request-with-list-and-first-datablock   [5] IMPLICIT Set-Request-With-List-And-First-Datablock
}

Set-Response ::= CHOICE
{
set-response-normal                     [1] IMPLICIT Set-Response-Normal,
set-response-datablock                  [2] IMPLICIT Set-Response-Datablock,
set-response-last-datablock             [3] IMPLICIT Set-Response-Last-Datablock,
set-response-last-datablock-with-list   [4] IMPLICIT Set-Response-Last-Datablock-With-List,
set-response-with-list                  [5] IMPLICIT Set-Response-With-List
}
"""


@attr.s(auto_attribs=True)
class SetRequestNormal(AbstractXDlmsApdu):
    """
    Set-Request-Normal ::= SEQUENCE
    {
        invoke-id-and-priority          Invoke-Id-And-Priority,
        cosem-attribute-descriptor      Cosem-Attribute-Descriptor,
        access-selection                Selective-Access-Descriptor OPTIONAL,
        value                           Data
    }
    """

    TAG: ClassVar[int] = 193
    RESPONSE_TYPE: ClassVar[enums.SetRequestType] = enums.SetRequestType.NORMAL
    cosem_attribute: cosem.CosemAttribute = attr.ib(
        validator=attr.validators.instance_of(cosem.CosemAttribute)
    )
    data: bytes = attr.ib(validator=attr.validators.instance_of(bytes))
    access_selection: Optional[Any] = attr.ib(default=None)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for SetRequest is not correct. Got {tag}, should be {cls.TAG}"
            )

        type_choice = enums.SetRequestType(data.pop(0))
        if type_choice is not enums.SetRequestType.NORMAL:
            raise ValueError("The type of the SetRequest is not for a SetRequestNormal")

        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        cosem_attribute = cosem.CosemAttribute.from_bytes(data[:9])
        data = data[9:]

        has_access_selection = bool(data.pop(0))
        if has_access_selection:
            raise NotImplementedError("Selective access on SET is not implemented")
        else:
            access_selection = None

        return cls(
            cosem_attribute=cosem_attribute,
            data=bytes(data),
            access_selection=access_selection,
            invoke_id_and_priority=invoke_id_and_priority,
        )

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(self.cosem_attribute.to_bytes())
        if self.access_selection:
            out.extend(b"\x01")
            out.extend(self.access_selection.to_bytes())
        else:
            out.extend(b"\x00")
        out.extend(self.data)
        return bytes(out)


@attr.s(auto_attribs=True)
class SetRequestWithFirstBlock:
    """
    Set-Request-With-First-Datablock ::= SEQUENCE
    {
    invoke-id-and-priority      Invoke-Id-And-Priority,
    cosem-attribute-descriptor  Cosem-Attribute-Descriptor,
    access-selection            [0] IMPLICIT Selective-Access-Descriptor OPTIONAL,
    datablock                   DataBlock-SA
    }

    DataBlock-SA ::= SEQUENCE  -- SA == DataBlock for the SET-request, ACTION-request and ACTION-response
    {
    last-block      BOOLEAN,
    block-number    Unsigned32,
    raw-data        OCTET STRING
    }
    """

    ...


@attr.s(auto_attribs=True)
class SetRequestWithBlock:
    """
    Set-Request-With-Datablock ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    datablock               DataBlock-SA
    }

    DataBlock-SA ::= SEQUENCE  -- SA == DataBlock for the SET-request, ACTION-request and ACTION-response
    {
    last-block      BOOLEAN,
    block-number    Unsigned32,
    raw-data        OCTET STRING
    }
    """

    ...


@attr.s(auto_attribs=True)
class SetRequestWithList:
    """
    Set-Request-With-List ::= SEQUENCE
    {
    invoke-id-and-priority      Invoke-Id-And-Priority,
    attribute-descriptor-list   SEQUENCE OF Cosem-Attribute-Descriptor-With-Selection,
    value-list                  SEQUENCE OF Data
    }

    Cosem-Attribute-Descriptor-With-Selection ::= SEQUENCE
    {
    cosem-attribute-descriptor   Cosem-Attribute-Descriptor,
    access-selection             Selective-Access-Descriptor OPTIONAL
    }
    """

    ...


@attr.s(auto_attribs=True)
class SetRequestWithListFirstBlock:
    """
    Set-Request-With-List-And-First-Datablock ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    attribute-descriptor-list  SEQUENCE OF Cosem-Attribute-Descriptor-With-Selection,
    datablock DataBlock-SA
    }

    Cosem-Attribute-Descriptor-With-Selection ::= SEQUENCE
    {
    cosem-attribute-descriptor   Cosem-Attribute-Descriptor,
    access-selection             Selective-Access-Descriptor OPTIONAL
    }

    DataBlock-SA ::= SEQUENCE  -- SA == DataBlock for the SET-request, ACTION-request and ACTION-response
    {
    last-block      BOOLEAN,
    block-number    Unsigned32,
    raw-data        OCTET STRING
    }

    """

    ...


@attr.s(auto_attribs=True)
class SetRequestFactory:
    """
    The factory will parse the SetRequest and return either a SetRequest type class
    """

    TAG: ClassVar[int] = 193

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != SetRequestFactory.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be "
                f"{SetRequestFactory.TAG}"
            )
        request_type = enums.SetRequestType(data.pop(0))
        if request_type == enums.SetRequestType.NORMAL:
            return SetRequestNormal.from_bytes(source_bytes)

        else:
            raise NotImplementedError("Only SetRequestNormal implemented")


@attr.s(auto_attribs=True)
class SetResponseNormal(AbstractXDlmsApdu):
    """
    Set-Response-Normal ::= SEQUENCE
    {
        invoke-id-and-priority Invoke-Id-And-Priority,
        result  Data-Access-Result
    }
    """

    TAG: ClassVar[int] = 197
    RESPONSE_TYPE: ClassVar[enums.SetResponseType] = enums.SetResponseType.NORMAL
    result: enums.DataAccessResult = attr.ib(
        validator=attr.validators.instance_of(enums.DataAccessResult)
    )
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for SetResponse is not correct. Got {tag}, should be {cls.TAG}"
            )

        type_choice = enums.SetResponseType(data.pop(0))
        if type_choice is not enums.SetResponseType.NORMAL:
            raise ValueError(
                "The type of the SetResponse is not for a SetResponseNormal"
            )

        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )

        result = enums.DataAccessResult(data.pop(0))

        return cls(result=result, invoke_id_and_priority=invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(self.result.value)
        return bytes(out)


@attr.s(auto_attribs=True)
class SetResponseWithBlock:
    """
    Set-Response-Datablock ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    block-number            Unsigned32
    }
    """

    ...


@attr.s(auto_attribs=True)
class SetResponseLastBlock:
    """
    Set-Response-Last-Datablock ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    result                  Data-Access-Result,
    block-number            Unsigned32
    }
    """

    ...


@attr.s(auto_attribs=True)
class SetResponseLastBlockWithList:
    """
    Set-Response-Last-Datablock-With-List ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    result                  SEQUENCE OF Data-Access-Result,
    block-number            Unsigned32
    }
    """

    ...


@attr.s(auto_attribs=True)
class SetResponseWithList:
    """
    Set-Response-With-List ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    result                  SEQUENCE OF Data-Access-Result,
    }
    """

    ...


@attr.s(auto_attribs=True)
class SetResponseFactory:
    """
    The factory will parse the SetResponse and return a SetResponse type class
    """

    TAG: ClassVar[int] = 197

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != SetResponseFactory.TAG:
            raise ValueError(
                f"Tag for Set response is not correct. Got {tag}, should be "
                f"{SetResponseFactory.TAG}"
            )
        request_type = enums.SetResponseType(data.pop(0))
        if request_type == enums.SetResponseType.NORMAL:
            return SetResponseNormal.from_bytes(source_bytes)

        else:
            raise NotImplementedError("Only SetResponseNormal implemented")



from typing import *

import attr

from dlms_cosem import ber


@attr.s(auto_attribs=True)
class UserInformation:
    """
    UserInformation holds InitiateRequests for AARQ and InitiateResponse for AARE.
    In case of error it can hold an ConformedServiceErrorAPDU in the AARE.
    In case of encryption the user-information holds ciphered APDUs. Either global-ciper
    or dedicated-cipher #TODO: is dedicated reasonable since no association has started.

    All the APDUs held by the user information is encoded in X-ADR but the AARQ/AARE are
    encoded in BER. To be able to make it distinct the content of the endoed XDLMS-APDU
    is encoded as an OctetString in BER.

    """

    tag = b"\x04"  # is encoded as an octetstring

    content: Any
    # Union[
    # xdlms.InitiateRequestApdu,
    # xdlms.InitiateResponseApdu,
    # xdlms.ConfirmedServiceErrorApdu,
    # xdlms.GlobalCipherInitiateRequest
    # ]

    @classmethod
    def from_bytes(cls, _bytes):
        from dlms_cosem.protocol import xdlms

        tag, length, data = ber.BER.decode(_bytes)
        if tag != UserInformation.tag:
            raise ValueError(
                f"The tag for UserInformation data should be 0x04" f"not {tag!r}"
            )

        if data[0] == 1:
            return cls(content=xdlms.InitiateRequest.from_bytes(data))
        elif data[0] == 8:
            return cls(content=xdlms.InitiateResponse.from_bytes(data))
        elif data[0] == 14:
            return cls(content=xdlms.ConfirmedServiceError.from_bytes(data))
        elif data[0] == 33:
            return cls(content=xdlms.GlobalCipherInitiateRequest.from_bytes(data))
        elif data[0] == 40:
            return cls(content=xdlms.GlobalCipherInitiateResponse.from_bytes(data))
        else:
            raise ValueError(
                f"Not able to find a proper data tag in UserInformation. Got {data[0]}"
            )

    def to_bytes(self):
        return ber.BER.encode(self.tag, self.content.to_bytes())



import abc
from typing import *

import attr

from dlms_cosem import enumerations
from dlms_cosem.ber import BER


class AbstractAcseApdu(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def from_bytes(cls, source_bytes: bytes):
        raise NotImplementedError("")

    @abc.abstractmethod
    def to_bytes(self) -> bytes:
        raise NotImplementedError("")


@attr.s(auto_attribs=True)
class DLMSObjectIdentifier:
    """
    The DLMS Association has been assigned a prefix for all of its OBJECT
    IDENDIFIERS
    """

    TAG: ClassVar[bytes] = b"\x06"
    PREFIX: ClassVar[bytes] = b"\x60\x85\x74\x05\x08"


@attr.s(auto_attribs=True)
class AppContextName(DLMSObjectIdentifier):
    """
    This defines how to reference objects in the meter and if ciphered APDU:s
    are allowed.
    """

    # TODO: Can this be a bit more generalized??
    app_context: ClassVar[int] = 1

    valid_context_ids: ClassVar[List[int]] = [1, 2, 3, 4]

    logical_name_refs: bool = attr.ib(default=True)
    ciphered_apdus: bool = attr.ib(default=True)

    @property
    def context_id(self) -> int:
        if self.logical_name_refs and not self.ciphered_apdus:
            return 1
        elif not self.logical_name_refs and not self.ciphered_apdus:
            return 2
        elif self.logical_name_refs and self.ciphered_apdus:
            return 3
        elif not self.logical_name_refs and self.ciphered_apdus:
            return 4
        else:
            raise ValueError(
                "Combination of logical name ref and " "ciphered apdus not possible"
            )

    @classmethod
    def from_bytes(cls, _bytes):
        tag, length, data = BER.decode(_bytes)

        if tag != DLMSObjectIdentifier.TAG:
            raise ValueError(
                f"Tag of {tag} is not a valid tag for " f"ObjectIdentifiers"
            )

        context_id = data[-1]
        if context_id not in AppContextName.valid_context_ids:
            raise ValueError(f"context_id of {context_id} is not valid")

        total_prefix = bytes(data[:-1])
        if total_prefix != (
            DLMSObjectIdentifier.PREFIX + bytes([AppContextName.app_context])
        ):
            raise ValueError(
                f"Static part of object id it is not correct"
                f" according to DLMS: {total_prefix}"
            )
        settings_dict = AppContextName.get_settings_by_context_id(context_id)
        return cls(**settings_dict)

    def to_bytes(self):
        total_data = self.PREFIX + bytes([self.app_context, self.context_id])
        return BER.encode(self.TAG, total_data)

    @staticmethod
    def get_settings_by_context_id(context_id):
        settings_dict = {
            1: {"logical_name_refs": True, "ciphered_apdus": False},
            2: {"logical_name_refs": False, "ciphered_apdus": False},
            3: {"logical_name_refs": True, "ciphered_apdus": True},
            4: {"logical_name_refs": False, "ciphered_apdus": True},
        }
        return settings_dict.get(context_id)


@attr.s(auto_attribs=True)
class MechanismName(DLMSObjectIdentifier):
    app_context: ClassVar[int] = 2

    mechanism: enumerations.AuthenticationMechanism

    @classmethod
    def from_bytes(cls, _bytes: bytes):
        """
        Apparently the data in mechanism name is not encoded in BER.
        """

        mechanism_id: int = _bytes[-1]

        total_prefix = bytes(_bytes[:-1])
        if total_prefix != (
            DLMSObjectIdentifier.PREFIX + bytes([MechanismName.app_context])
        ):
            raise ValueError(
                f"Static part of object id it is not correct"
                f" according to DLMS: {total_prefix!r}"
            )

        return cls(mechanism=enumerations.AuthenticationMechanism(mechanism_id))

    def to_bytes(self):
        total_data = self.PREFIX + bytes([self.app_context, self.mechanism.value])
        return total_data


def validate_password_type(instance, attribute, value):

    if value not in AuthenticationValue.allowed_password_types:
        raise ValueError(f"{value} is not a valid auth value type")


@attr.s(auto_attribs=True)
class AuthenticationValue:
    """
    Holds "password" in the AARQ and AARE
    Can either hold a charstring or a bitstring
    """

    password: bytes = attr.ib(default=b"")
    password_type: str = attr.ib(default="chars", validator=[validate_password_type])
    allowed_password_types: ClassVar[List[str]] = ["chars", "bits"]

    @classmethod
    def from_bytes(cls, _bytes):
        tag, length, data = BER.decode(_bytes)
        if tag == b"\x80":
            password_type = "chars"
        elif tag == b"\x81":
            password_type = "bits"
        else:
            raise ValueError(f"Tag {tag} is not vaild for password")

        return cls(password=data, password_type=password_type)

    def to_bytes(self):
        if self.password_type == "chars":
            return BER.encode(0x80, self.password)
        elif self.password_type == "bits":
            return BER.encode(0x81, self.password)


@attr.s(auto_attribs=True)
class AuthFunctionalUnit:
    """
    Consists of 2 bytes. First byte encodes the number of unused bytes in
    the second byte.
    So really you just need to set the last bit to 0 to use authentication.
    In the green book they use the 0x07 as first byte and 0x80 as last byte.
    We will use this to not make it hard to look up.
    It is a bit weirdly defined in the Green Book. I interpret is as if the data
    exists it is the functional unit 0 (authentication). In examples in the
    Green Book they set 0x070x80 as exists.
    """

    authentication: bool = attr.ib(default=False)

    @classmethod
    def from_bytes(cls, _bytes):
        if len(_bytes) != 2:
            raise ValueError(
                f"Authentication Functional Unit data should by 2 "
                f"bytes. Got: {_bytes}"
            )
        last_byte = bool(_bytes[-1])
        return cls(authentication=last_byte)

    def to_bytes(self):
        if self.authentication:
            return b"\x07\x80"
        else:
            # when not using authentication this the sender-acse-requirements
            # should not be in the data.
            return None



import pytest

from dlms_cosem import enumerations
from dlms_cosem.protocol.acse import ApplicationAssociationRequest, UserInformation
from dlms_cosem.protocol.xdlms import (
    Conformance,
    GlobalCipherInitiateResponse,
    InitiateRequest,
)

# Example encodings from DLMS Green Book v10: page 444


class TestParseAARQ:
    def test_parse_no_ciphering_no_sercurity(self):
        data = b"`\x1d\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\xbe\x10\x04\x0e\x01\x00\x00\x00\x06_\x1f\x04\x00\x00~\x1f\x04\xb0"
        aarq = ApplicationAssociationRequest.from_bytes(data)
        print(aarq)
        assert not aarq.ciphered
        assert aarq.authentication is None

    def test_parse_no_ciphering_low_security(self):
        data = b"`6\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\x8a\x02\x07\x80\x8b\x07`\x85t\x05\x08\x02\x01\xac\n\x80\x0812345678\xbe\x10\x04\x0e\x01\x00\x00\x00\x06_\x1f\x04\x00\x00~\x1f\x04\xb0"
        aarq = ApplicationAssociationRequest.from_bytes(data)
        print(aarq)
        assert not aarq.ciphered
        assert aarq.authentication == enumerations.AuthenticationMechanism.LLS
        # Password is used in LLS
        assert aarq.authentication_value is not None

    def test_parse_no_ciphering_high_security(self):
        data = b"`6\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\x8a\x02\x07\x80\x8b\x07`\x85t\x05\x08\x02\x05\xac\n\x80\x08K56iVagY\xbe\x10\x04\x0e\x01\x00\x00\x00\x06_\x1f\x04\x00\x00~\x1f\x04\xb0"
        aarq = ApplicationAssociationRequest.from_bytes(data)
        print(aarq)
        assert not aarq.ciphered
        assert aarq.authentication == enumerations.AuthenticationMechanism.HLS_GMAC
        # Password is used in LLS
        assert aarq.authentication_value is not None

    def test_parse_ciphered_low_security(self):
        data = b'`f\xa1\t\x06\x07`\x85t\x05\x08\x01\x03\xa6\n\x04\x08MMM\x00\x00\xbcaN\x8a\x02\x07\x80\x8b\x07`\x85t\x05\x08\x02\x01\xac\n\x80\x0812345678\xbe4\x042!00\x01#Eg\x80\x13\x02\xff\x8axt\x13=AL\xed%\xb4%4\xd2\x8d\xb0\x04w `k\x17[\xd5"\x11\xbehA\xdb M9\xeeo\xdb\x8e5hU'
        aarq = ApplicationAssociationRequest.from_bytes(data)
        print(aarq)
        assert aarq.ciphered
        assert aarq.authentication == enumerations.AuthenticationMechanism.LLS
        # you need to set a system title when ciphering
        assert aarq.system_title is not None
        # Password is used in LLS
        assert aarq.authentication_value is not None

    def test_parse_ciphered_low_security2(self):
        data = bytes.fromhex(
            "6066a109060760857405080103a60a04084D4D4D0000BC614E8a0207808b0760857405080201ac0a80083132333435363738be34043221303001234567801302FF8A7874133D414CED25B42534D28DB0047720606B175BD52211BE6841DB204D39EE6FDB8E356855"
        )
        aarq = ApplicationAssociationRequest.from_bytes(data)
        print(aarq)
        assert aarq.ciphered
        assert aarq.authentication == enumerations.AuthenticationMechanism.LLS
        # you need to set a system title when ciphering
        assert aarq.system_title == bytes.fromhex("4D4D4D0000BC614E")
        # Password is used in LLS
        assert aarq.authentication_value == b"12345678"
        assert aarq.to_bytes() == data

    def test_hls(self):
        data = bytes.fromhex(
            "6036A1090607608574050801018A0207808B0760857405080202AC0A80083132333435363738BE10040E01000000065F1F0400007E1FFFFF"
        )
        aarq = ApplicationAssociationRequest.from_bytes(data)
        assert aarq.authentication
        assert aarq.user_information
        assert isinstance(aarq.user_information.content, InitiateRequest)
        assert aarq.authentication_value is not None


class TestEncodeAARE:
    def test_decode(self):
        resulting_bytes = bytes.fromhex(
            "601DA109060760857405080101BE10040E01000000065F1F0400001E1DFFFF"
        )
        aarq = ApplicationAssociationRequest(
            ciphered=False,
            system_title=None,
            public_cert=None,
            authentication=None,
            authentication_value=None,
            user_information=UserInformation(
                content=InitiateRequest(
                    proposed_conformance=Conformance(
                        general_protection=False,
                        general_block_transfer=False,
                        delta_value_encoding=False,
                        attribute_0_supported_with_set=False,
                        priority_management_supported=False,
                        attribute_0_supported_with_get=False,
                        block_transfer_with_get_or_read=True,
                        block_transfer_with_set_or_write=True,
                        block_transfer_with_action=True,
                        multiple_references=True,
                        data_notification=False,
                        access=False,
                        get=True,
                        set=True,
                        selective_access=True,
                        event_notification=False,
                        action=True,
                    ),
                    proposed_quality_of_service=0,
                    client_max_receive_pdu_size=65535,
                    proposed_dlms_version_number=6,
                    response_allowed=True,
                    dedicated_key=None,
                )
            ),
        )

        assert aarq.to_bytes() == resulting_bytes

    def test_parse_no_ciphering_no_sercurity(self):
        data = b"`\x1d\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\xbe\x10\x04\x0e\x01\x00\x00\x00\x06_\x1f\x04\x00\x00~\x1f\x04\xb0"
        aarq = ApplicationAssociationRequest(
            system_title=None,
            public_cert=None,
            authentication=None,
            ciphered=False,
            authentication_value=None,
            user_information=UserInformation(
                content=InitiateRequest(
                    proposed_conformance=Conformance(
                        general_protection=False,
                        general_block_transfer=False,
                        delta_value_encoding=False,
                        attribute_0_supported_with_set=False,
                        priority_management_supported=True,
                        attribute_0_supported_with_get=True,
                        block_transfer_with_get_or_read=True,
                        block_transfer_with_set_or_write=True,
                        block_transfer_with_action=True,
                        multiple_references=True,
                        data_notification=False,
                        access=False,
                        get=True,
                        set=True,
                        selective_access=True,
                        event_notification=True,
                        action=True,
                    ),
                    proposed_quality_of_service=0,
                    client_max_receive_pdu_size=1200,
                    proposed_dlms_version_number=6,
                    response_allowed=True,
                    dedicated_key=None,
                )
            ),
            calling_ae_invocation_identifier=None,
            called_ap_title=None,
            called_ae_qualifier=None,
            called_ap_invocation_identifier=None,
            called_ae_invocation_identifier=None,
            calling_ap_invocation_identifier=None,
            implementation_information=None,
        )
        assert aarq.to_bytes() == data

    def test_encode_no_ciphering_high_security(self):
        data = b"`6\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\x8a\x02\x07\x80\x8b\x07`\x85t\x05\x08\x02\x05\xac\n\x80\x08K56iVagY\xbe\x10\x04\x0e\x01\x00\x00\x00\x06_\x1f\x04\x00\x00~\x1f\x04\xb0"
        aarq = ApplicationAssociationRequest(
            system_title=None,
            public_cert=None,
            authentication=enumerations.AuthenticationMechanism.HLS_GMAC,
            ciphered=False,
            authentication_value=b"K56iVagY",
            user_information=UserInformation(
                content=InitiateRequest(
                    proposed_conformance=Conformance(
                        general_protection=False,
                        general_block_transfer=False,
                        delta_value_encoding=False,
                        attribute_0_supported_with_set=False,
                        priority_management_supported=True,
                        attribute_0_supported_with_get=True,
                        block_transfer_with_get_or_read=True,
                        block_transfer_with_set_or_write=True,
                        block_transfer_with_action=True,
                        multiple_references=True,
                        data_notification=False,
                        access=False,
                        get=True,
                        set=True,
                        selective_access=True,
                        event_notification=True,
                        action=True,
                    ),
                    proposed_quality_of_service=0,
                    client_max_receive_pdu_size=1200,
                    proposed_dlms_version_number=6,
                    response_allowed=True,
                    dedicated_key=None,
                )
            ),
            calling_ae_invocation_identifier=None,
            called_ap_title=None,
            called_ae_qualifier=None,
            called_ap_invocation_identifier=None,
            called_ae_invocation_identifier=None,
            calling_ap_invocation_identifier=None,
            implementation_information=None,
        )

        assert aarq.to_bytes() == data

    def test_encode_no_ciphering_low_security(self):
        data = b"`6\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\x8a\x02\x07\x80\x8b\x07`\x85t\x05\x08\x02\x01\xac\n\x80\x0812345678\xbe\x10\x04\x0e\x01\x00\x00\x00\x06_\x1f\x04\x00\x00~\x1f\x04\xb0"
        aarq = ApplicationAssociationRequest(
            system_title=None,
            public_cert=None,
            authentication=enumerations.AuthenticationMechanism.LLS,
            ciphered=False,
            authentication_value=b"12345678",
            user_information=UserInformation(
                content=InitiateRequest(
                    proposed_conformance=Conformance(
                        general_protection=False,
                        general_block_transfer=False,
                        delta_value_encoding=False,
                        attribute_0_supported_with_set=False,
                        priority_management_supported=True,
                        attribute_0_supported_with_get=True,
                        block_transfer_with_get_or_read=True,
                        block_transfer_with_set_or_write=True,
                        block_transfer_with_action=True,
                        multiple_references=True,
                        data_notification=False,
                        access=False,
                        get=True,
                        set=True,
                        selective_access=True,
                        event_notification=True,
                        action=True,
                    ),
                    proposed_quality_of_service=0,
                    client_max_receive_pdu_size=1200,
                    proposed_dlms_version_number=6,
                    response_allowed=True,
                    dedicated_key=None,
                )
            ),
            calling_ae_invocation_identifier=None,
            called_ap_title=None,
            called_ae_qualifier=None,
            called_ap_invocation_identifier=None,
            called_ae_invocation_identifier=None,
            calling_ap_invocation_identifier=None,
            implementation_information=None,
        )
        assert aarq.to_bytes() == data



from typing import *

import attr

from dlms_cosem import enumerations
from dlms_cosem.ber import BER
from dlms_cosem.protocol import xdlms
from dlms_cosem.protocol.acse import base as acse_base
from dlms_cosem.protocol.acse.user_information import UserInformation


def user_information_holds_initiate_request(
    instance, attribute, value: UserInformation
):
    if not isinstance(
        value.content, (xdlms.InitiateRequest, xdlms.GlobalCipherInitiateRequest)
    ):
        raise ValueError(
            f"ApplicationAssociationRequestApdu.user_information should "
            f"only hold a UserInformation where .content is a "
            f"InitiateRequestApdu or GlobalCihperInitiateRequest. Got {value.content.__class__.__name__}"
        )


def aarq_should_set_authenticated(
    mechanism: Optional[enumerations.AuthenticationMechanism],
):
    """
    * If Lowest Level Scurity (None) is used it shall not be present.
    * If Low Level Security (LLS) is used it should be present in request and may
        be present in response (AARE) and indicate authentication (bit0 = 1)
    * If High Level Security (HLS) is used it should be present in both request and
        response (AARE) and indicate authentication (bit0 = 1)
    """
    if not mechanism:
        return False

    if mechanism == enumerations.AuthenticationMechanism.NONE:
        return False

    return True


@attr.s(auto_attribs=True)
class ApplicationAssociationRequest:
    """
    Application Association Request ( AARQ ) is used for starting an Application
    Association with a DLMS server (meter).

    It is encoded in BER. This is due to many of
    the field are variable in length and it is hard to know the length of each field in
    advance.

    The user information field holds an InitiateRequestApdu encoded in X-ADR wrapped
    in BER encoded OCTET_STRING.

    If sender_acse_requirements is present authentication is used.

    mechanism_name and calling_authentication_value should only be present if
    authentication is used.

    AP = Application Process
    AE = Application Entity
    ACSE = Association Control Service Element

    :parameter ciphered: Sets the AppContextName to indicate ciphered apdus.

    :parameter client_system_title:  Is transferred in the `calling_ap_title` part of
        AARQ as defined in the DLMS ASN1 Specs.If the `application_context_name` uses
        ciphering `calling_ap_title` should contain the clients system title. Also if the
        proposed HLS (High Level Security) auth mechanism uses the system title it
        should be present.

    :parameter client_public_cert:  Is transferred in the `calling_ae_qualifier`part of
        the AARQ as defiened in the DLMS ASN1 Specs.If the `application_context_name`
        indicates the use of ciphered APDUs the `calling_ae_qualifier` may hold the
        public digital signature key certificate of the client.

    :parameter authentication: Defines the type of authentication that is used.

    :parameter authentication_value: Is tranferred in the `calling_authentication_value`
        part of the AARQ as defined in the DLMS ASN1 specs.Shall only be present if
        `sender_acse_requirements` indicates authentication. It holds the client
        authentication value appropriat to the selected `mechanism_name`.

    :parameter user_information: Holds the proposed xDLMS context in the form of a
        `InitiateRequestApdu` encoded as a BER OctetString.
        If the `InitiateRequestApdu` contains a dedicated_key it should be
        authenticated and encrypted using the AES-GCM, global encryption key and the
        authentication key (if in use).
        Even if there is no dedicated key the `InitiateRequestApdu` should be protected
        as above if there is need to protect the RLRQ

    :parameter calling_ae_invocation_identifier: When used it holds the user_id of the
        client. User ideentification is an optional feature, `Blue Bool 4.4.2`.
        Not abvailable on presetablished associations.
        The server (meter) should hold a users list and if the user_id is not present
        in the meter the assiciation request is rejected.


    :parameter called_ap_title: Usage not defined in DLMS green book.
        Usage could be defined by meter manufacturer

    :parameter called_ae_qualifier: Usage not defined in DLMS green book.
        Usage could be defined by meter manufacturer

    :parameter called_ap_invocation_identfier: Usage not defined in DLMS green book.
        Usage could be defined by meter manufacturer

    :parameter called_ae_invocation_indentifier: Usage not defined in DLMS green book.
        Usage could be defined by meter manufacturer

    :parameter calling_ap_invocation_identifier: Usage not defined in DLMS green book.
        Usage could be defined by meter manufacturer



    :parameter implementation_information: Usage not defined in DLMS green book.
        Usage could be defined by meter manufacturer

    }
    """

    TAG: ClassVar[int] = 0x60  # Application 0 = 60H = 96
    PARSE_TAGS = {
        0x80: ("protocol_version", None),  # Context specific, constructed? 0
        0xA1: ("application_context_name", acse_base.AppContextName),
        # Context specific, constructed 1
        162: ("called_ap_title", None),
        163: ("called_ae_qualifier", None),
        164: ("called_ap_invocation_identifier", None),
        165: ("called_ae_invocation_identifier", None),
        166: ("calling_ap_title", None),
        167: ("calling_ae_qualifier", None),
        168: ("calling_ap_invocation_identifier", None),
        169: ("calling_ae_invocation_identifier", None),
        0x8A: ("sender_acse_requirements", acse_base.AuthFunctionalUnit),
        0x8B: ("mechanism_name", acse_base.MechanismName),
        0xAC: ("calling_authentication_value", acse_base.AuthenticationValue),
        0xBD: ("implementation_information", None),
        0xBE: ("user_information", UserInformation),  # Context specific, constructed 30
    }

    user_information: UserInformation = attr.ib(
        validator=[user_information_holds_initiate_request]
    )
    system_title: Optional[bytes] = attr.ib(default=None)
    public_cert: Optional[bytes] = attr.ib(default=None)
    authentication: Optional[enumerations.AuthenticationMechanism] = attr.ib(
        default=None
    )
    ciphered: bool = attr.ib(default=False)
    # TODO: Can we rename this to password? Would be nice to pass it as bytes or str.
    authentication_value: Optional[bytes] = attr.ib(default=None)
    # TODO: validate that a ciphered InitiateReqest is used when ciphering is True.
    calling_ae_invocation_identifier: Optional[bytes] = attr.ib(default=None)

    # Not really used
    # TODO: Should we keep them?
    called_ap_title: Optional[bytes] = attr.ib(default=None)
    called_ae_qualifier: Optional[bytes] = attr.ib(default=None)
    called_ap_invocation_identifier: Optional[bytes] = attr.ib(default=None)
    called_ae_invocation_identifier: Optional[bytes] = attr.ib(default=None)
    calling_ap_invocation_identifier: Optional[bytes] = attr.ib(default=None)
    implementation_information: Optional[bytes] = attr.ib(default=None)

    @property
    def sender_acse_requirements(self) -> Optional[acse_base.AuthFunctionalUnit]:
        """
        Is only sent if the AuthFunctionalUnit needs to indicate authentication.
        """
        if aarq_should_set_authenticated(self.authentication):
            return acse_base.AuthFunctionalUnit(True)

        return None

    @property
    def mechanism_name(self) -> Optional[acse_base.MechanismName]:
        """
        The mechanism_name field should only be present if the AuthFunctionalUnit
        indicates authenticated.
        :return:
        """
        if (
            self.sender_acse_requirements is not None
            and self.authentication is not None
        ):
            return acse_base.MechanismName(mechanism=self.authentication)

        return None

    @property
    def application_context_name(self) -> acse_base.AppContextName:
        """
        Always use logical name references.
        """
        if self.ciphered:
            return acse_base.AppContextName(logical_name_refs=True, ciphered_apdus=True)
        else:
            return acse_base.AppContextName(
                logical_name_refs=True, ciphered_apdus=False
            )

    @property
    def protocol_version(self) -> int:
        return 0

    @classmethod
    def from_bytes(cls, aarq_bytes):
        # put it in a bytearray to be able to pop.
        aarq_data = bytearray(aarq_bytes)

        aarq_tag = aarq_data.pop(0)
        if not aarq_tag == cls.TAG:
            raise ValueError("Bytes are not an AARQ APDU. TAg is not int(96)")

        aarq_length = aarq_data.pop(0)

        if not len(aarq_data) == aarq_length:
            raise ValueError(
                "The APDU Data lenght does not correspond " "to length byte"
            )

        # Assumes that the protocol-version is 1 and we don't need to decode it

        # Decode the AARQ  data
        object_dict = dict()
        # use the data in tags to go through the bytes and create objects.
        while True:
            # TODO: this does not take into account when defining objects in dict and not using them.
            object_tag = aarq_data.pop(0)
            object_desc = ApplicationAssociationRequest.PARSE_TAGS.get(object_tag, None)
            if object_desc is None:
                raise ValueError(
                    f"Could not find object with tag {object_tag} "
                    f"in AARQ definition"
                )

            object_length = aarq_data.pop(0)
            object_data = bytes(aarq_data[:object_length])
            aarq_data = aarq_data[object_length:]

            object_name = object_desc[0]
            object_class: Any = object_desc[1]

            if object_class is not None:
                object_data = object_class.from_bytes(object_data)

            object_dict[object_name] = object_data

            if len(aarq_data) <= 0:
                break

        protocol_version: Optional[int] = object_dict.pop("protocol_version", None)
        if protocol_version is not None:
            if protocol_version != 0:
                raise ValueError("Parsed a protocol version that is not 0")

        application_context_name: acse_base.AppContextName = object_dict.pop(
            "application_context_name"
        )
        object_dict["ciphered"] = application_context_name.ciphered_apdus
        if not application_context_name.logical_name_refs:
            raise ValueError("Parsed a AARQ that uses Short Name Referencing!")

        sender_acse_requirements: Optional[
            acse_base.AuthFunctionalUnit
        ] = object_dict.pop("sender_acse_requirements", None)

        mechanism_name: Optional[acse_base.MechanismName] = object_dict.pop(
            "mechanism_name", None
        )

        if sender_acse_requirements and mechanism_name:
            object_dict["authentication"] = mechanism_name.mechanism

        # rename some elements
        client_system_title = object_dict.pop("calling_ap_title", None)
        if client_system_title:
            # it is ber encoded universal tag ocetctring. simple handling
            object_dict["system_title"] = client_system_title[2:]
        else:
            object_dict["system_title"] = None

        client_public_cert = object_dict.pop("calling_ae_qualifier", None)
        if client_public_cert:
            # it is ber encoded universal tag ocetctring. simple handling
            object_dict["public_cert"] = client_public_cert[2:]
        else:
            object_dict["public_cert"] = None

        auth_value: Optional[acse_base.AuthenticationValue] = object_dict.pop(
            "calling_authentication_value", None
        )
        if auth_value:
            object_dict["authentication_value"] = auth_value.password
        else:
            object_dict["authentication_value"] = None

        return cls(**object_dict)

    def to_bytes(self):
        aarq_data = bytearray()
        # There is no need to encode the version since it is always v1. (int 0)
        if self.application_context_name is not None:
            aarq_data.extend(BER.encode(161, self.application_context_name.to_bytes()))
        if self.called_ap_title is not None:
            aarq_data.extend(BER.encode(162, self.called_ap_title))
        if self.called_ae_qualifier is not None:
            aarq_data.extend(BER.encode(163, self.called_ae_qualifier))
        if self.called_ap_invocation_identifier is not None:
            aarq_data.extend(BER.encode(164, self.called_ap_invocation_identifier))
        if self.called_ae_invocation_identifier is not None:
            aarq_data.extend(BER.encode(165, self.called_ae_invocation_identifier))
        if self.system_title is not None:
            aarq_data.extend(BER.encode(166, BER.encode(4, self.system_title)))
        if self.public_cert is not None:
            aarq_data.extend(BER.encode(167, BER.encode(4, self.public_cert)))
        if self.calling_ap_invocation_identifier is not None:
            aarq_data.extend(BER.encode(168, self.calling_ap_invocation_identifier))
        if self.calling_ae_invocation_identifier is not None:
            aarq_data.extend(BER.encode(169, self.calling_ae_invocation_identifier))
        if self.sender_acse_requirements is not None:
            aarq_data.extend(BER.encode(0x8A, self.sender_acse_requirements.to_bytes()))
        if self.mechanism_name is not None:
            aarq_data.extend(BER.encode(0x8B, self.mechanism_name.to_bytes()))
        if self.authentication_value is not None:
            aarq_data.extend(
                BER.encode(
                    0xAC,
                    acse_base.AuthenticationValue(
                        password=self.authentication_value
                    ).to_bytes(),
                )
            )
        if self.implementation_information is not None:
            aarq_data.extend(BER.encode(0xBD, self.implementation_information))
        if self.user_information is not None:
            aarq_data.extend(BER.encode(0xBE, self.user_information.to_bytes()))

        return BER.encode(self.TAG, aarq_data)



import re
from typing import *

import attr

six_part = re.compile(
    "^(\\d{1,3}).(\\d{1,3}).(\\d{1,3}).(\\d{1,3}).(\\d{1,3}).(\\d{1,3})$"
)
five_part = re.compile("^(\\d{1,3}).(\\d{1,3}).(\\d{1,3}).(\\d{1,3}).(\\d{1,3})$")


def allowed_range_for_obis_code(instance, attribute, value: int):

    if 0 > value > 255:
        raise ValueError("An obis can only be between 0 - 255")


@attr.s(auto_attribs=True)
class Obis:

    """
    OBject Identification System defines codes for identification of commonly used
    data items in metering equipment.
    """

    a: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    b: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    c: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    d: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    e: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    f: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code],
        default=255,
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        if len(data) != 6:
            raise ValueError(
                f"Not enough data to parse OBIS. Need 6 bytes but got {len(data)}"
            )
        return cls(data[0], data[1], data[2], data[3], data[4], data[5])

    @classmethod
    def from_string(cls, obis_string: str) -> "Obis":
        """
        Parses a string as an OBIS code. Will accept with both the optinal 255 at the
        and and not. Any separator is allowed.
        """
        six_match = re.match(six_part, obis_string)
        if six_match:
            parts = six_match.groups()
            return cls(
                a=int(parts[0]),
                b=int(parts[1]),
                c=int(parts[2]),
                d=int(parts[3]),
                e=int(parts[4]),
            )
        five_match = re.match(five_part, obis_string)
        if five_match:
            parts = five_match.groups()
            return cls(
                a=int(parts[0]),
                b=int(parts[1]),
                c=int(parts[2]),
                d=int(parts[3]),
                e=int(parts[4]),
            )

        raise ValueError(f"{obis_string} is not a parsable OBIS string")

    def to_string(self, separator: Optional[str] = None) -> str:
        if separator:
            return (
                f"{self.a}{separator}{self.b}{separator}{self.c}{separator}{self.d}"
                f"{separator}{self.e}{separator}{self.f}"
            )
        else:
            return f"{self.a}-{self.b}:{self.c}.{self.d}.{self.e}.{self.f}"

    def to_bytes(self) -> bytes:
        return bytes(bytearray([self.a, self.b, self.c, self.d, self.e, self.f]))
